# -*- coding: utf-8 -*-
"""
    pint.unit
    ~~~~~~~~~

    Functions and classes related to unit definitions and conversions.

    :copyright: 2013 by Pint Authors, see AUTHORS for more details.
    :license: BSD, see LICENSE for more details.
"""

from __future__ import division, unicode_literals, print_function, absolute_import

import os
import math
import itertools
import functools
import operator
import pkg_resources
from decimal import Decimal
from contextlib import contextmanager, closing
from io import open, StringIO
from collections import defaultdict
from tokenize import untokenize, NUMBER, STRING, NAME, OP
from numbers import Number

from .context import Context, ContextChain
from .util import (logger, pi_theorem, solve_dependencies, ParserHelper,
                   string_preprocessor, find_connected_nodes,
                   find_shortest_path, UnitsContainer, _is_dim,
                   SharedRegistryObject, to_units_container)
from .compat import tokenizer, string_types, NUMERIC_TYPES, long_type, zip_longest
from .definitions import (Definition, UnitDefinition, PrefixDefinition,
                          DimensionDefinition)
from .converters import ScaleConverter
from .errors import (DimensionalityError, UndefinedUnitError,
                     DefinitionSyntaxError, RedefinitionError)

from .pint_eval import build_eval_tree
from . import systems


def _capture_till_end(ifile):
    context = []
    for no, line in ifile:
        line = line.strip()
        if line.startswith('@end'):
            break
        elif line.startswith('@'):
            raise DefinitionSyntaxError('cannot nest @ directives', lineno=no)
        context.append(line)
    return context


class _Unit(SharedRegistryObject):
    """Implements a class to describe a unit supporting math operations.

    :type units: UnitsContainer, str, Unit or Quantity.

    """

    #: Default formatting string.
    default_format = ''

    def __reduce__(self):
        from . import _build_unit
        return _build_unit, (self._units)

    def __new__(cls, units):
        inst = object.__new__(cls)
        if isinstance(units, (UnitsContainer, UnitDefinition)):
            inst._units = units
        elif isinstance(units, string_types):
            inst._units = inst._REGISTRY.parse_units(units)._units
        elif isinstance(units, _Unit):
            inst._units = units._units
        else:
            raise TypeError('units must be of type str, Unit or '
                            'UnitsContainer; not {0}.'.format(type(units)))

        inst.__used = False
        inst.__handling = None
        return inst

    @property
    def debug_used(self):
        return self.__used

    def __copy__(self):
        ret = self.__class__(self._units)
        ret.__used = self.__used
        return ret

    def __str__(self):
        return format(self)

    def __repr__(self):
        return "<Unit('{0}')>".format(self._units)

    def __format__(self, spec):
        spec = spec or self.default_format

        if '~' in spec:
            if self.dimensionless:
                return ''
            units = UnitsContainer(dict((self._REGISTRY._get_symbol(key),
                                         value)
                                   for key, value in self._units.items()))
            spec = spec.replace('~', '')
        else:
            units = self._units

        return '%s' % (format(units, spec))

    # IPython related code
    def _repr_html_(self):
        return self.__format__('H')

    def _repr_latex_(self):
        return "$" + self.__format__('L') + "$"

    @property
    def dimensionless(self):
        """Return true if the Unit is dimensionless.

        """
        return not bool(self.dimensionality)

    @property
    def dimensionality(self):
        """Unit's dimensionality (e.g. {length: 1, time: -1})

        """
        try:
            return self._dimensionality
        except AttributeError:
            dim = self._REGISTRY._get_dimensionality(self._units)
            self._dimensionality = dim

        return self._dimensionality

    def compatible_units(self, *contexts):
        if contexts:
            with self._REGISTRY.context(*contexts):
                return self._REGISTRY.get_compatible_units(self)

        return self._REGISTRY.get_compatible_units(self)

    def __mul__(self, other):
        if self._check(other):
            if isinstance(other, self.__class__):
                return self.__class__(self._units*other._units)
            else:
                return self._REGISTRY.Quantity(other._magnitude,
                                               self._units*other._units)

        return self._REGISTRY.Quantity(other, self._units)

    __rmul__ = __mul__

    def __truediv__(self, other):
        if self._check(other):
            if isinstance(other, self.__class__):
                return self.__class__(self._units/other._units)
            else:
                return self._REGISTRY.Quantity(1/other._magnitude,
                                           self._units/other._units)

        return self._REGISTRY.Quantity(1/other, self._units)

    def __rtruediv__(self, other):
        # As Unit and Quantity both handle truediv with each other rtruediv can
        # only be called for something different.
        if isinstance(other, NUMERIC_TYPES):
            return self._REGISTRY.Quantity(other, 1/self._units)
        elif isinstance(other, UnitsContainer):
            return self.__class__(other/self._units)
        else:
            return NotImplemented

    __div__ = __truediv__
    __rdiv__ = __rtruediv__

    def __pow__(self, other):
        if isinstance(other, NUMERIC_TYPES):
            return self.__class__(self._units**other)

        else:
            mess = 'Cannot power Unit by {}'.format(type(other))
            raise TypeError(mess)

    def __hash__(self):
        return self._units.__hash__()

    def __eq__(self, other):
        # We compare to the base class of Unit because each Unit class is
        # unique.
        if self._check(other):
            if isinstance(other, self.__class__):
                return self._units == other._units
            else:
                return other == self._REGISTRY.Quantity(1, self._units)

        elif isinstance(other, NUMERIC_TYPES):
            return other == self._REGISTRY.Quantity(1, self._units)

        else:
            return self._units == other

    def compare(self, other, op):
        self_q = self._REGISTRY.Quantity(1, self)

        if isinstance(other, NUMERIC_TYPES):
            return self_q.compare(other, op)
        elif isinstance(other, (_Unit, UnitsContainer, dict)):
            return self_q.compare(self._REGISTRY.Quantity(1, other), op)
        else:
            return NotImplemented

    __lt__ = lambda self, other: self.compare(other, op=operator.lt)
    __le__ = lambda self, other: self.compare(other, op=operator.le)
    __ge__ = lambda self, other: self.compare(other, op=operator.ge)
    __gt__ = lambda self, other: self.compare(other, op=operator.gt)

    def __int__(self):
        return int(self._REGISTRY.Quantity(1, self._units))

    def __long__(self):
        return long_type(self._REGISTRY.Quantity(1, self._units))

    def __float__(self):
        return float(self._REGISTRY.Quantity(1, self._units))

    def __complex__(self):
        return complex(self._REGISTRY.Quantity(1, self._units))

    __array_priority__ = 17

    def __array_prepare__(self, array, context=None):
        return 1

    def __array_wrap__(self, array, context=None):
        uf, objs, huh = context

        if uf.__name__ in ('true_divide', 'divide', 'floor_divide'):
            return self._REGISTRY.Quantity(array, 1/self._units)
        elif uf.__name__ in ('multiply',):
            return self._REGISTRY.Quantity(array, self._units)
        else:
            raise ValueError('Unsupproted operation for Unit')


class UnitRegistry(object):
    """The unit registry stores the definitions and relationships between
    units.

    :param filename: path of the units definition file to load.
                     Empty to load the default definition file.
                     None to leave the UnitRegistry empty.
    :param force_ndarray: convert any input, scalar or not to a numpy.ndarray.
    :param default_as_delta: In the context of a multiplication of units, interpret
                             non-multiplicative units as their *delta* counterparts.
    :autoconvert_offset_to_baseunit: If True converts offset units in quantites are
                                     converted to their base units in multiplicative
                                     context. If False no conversion happens.
    :param on_redefinition: action to take in case a unit is redefined.
                            'warn', 'raise', 'ignore'
    :type on_redefintion: str
    """

    def __init__(self, filename='', force_ndarray=False, default_as_delta=True,
                 autoconvert_offset_to_baseunit=False,
                 on_redefinition='warn', system=None):
        self.Unit = build_unit_class(self)
        self.Quantity = build_quantity_class(self, force_ndarray)
        self.Measurement = build_measurement_class(self, force_ndarray)

        #: Action to take in case a unit is redefined. 'warn', 'raise', 'ignore'
        self._on_redefinition = on_redefinition

        #: Map dimension name (string) to its definition (DimensionDefinition).
        self._dimensions = {}

        #: :type: systems.GSManager
        self._gsmanager = systems.GSManager()

        #: Map unit name (string) to its definition (UnitDefinition).
        #: Might contain prefixed units.
        self._units = {}

        #: Map unit name in lower case (string) to a set of unit names with the right case.
        #: Does not contain prefixed units.
        #: e.g: 'hz' - > set('Hz', )
        self._units_casei = defaultdict(set)

        #: Map prefix name (string) to its definition (PrefixDefinition).
        self._prefixes = {'': PrefixDefinition('', '', (), 1)}

        #: Map suffix name (string) to canonical , and unit alias to canonical unit name
        self._suffixes = {'': None, 's': ''}

        #: Map context name (string) or abbreviation to context.
        self._contexts = {}

        #: Stores active contexts.
        self._active_ctx = ContextChain()

        #: Maps dimensionality (UnitsContainer) to Units (str)
        self._dimensional_equivalents = dict()

        #: Maps dimensionality (UnitsContainer) to Dimensionality (UnitsContainer)
        self._root_units_cache = dict()

        #: Maps dimensionality (UnitsContainer) to Dimensionality (UnitsContainer)
        self._base_units_cache = dict()

        #: Maps dimensionality (UnitsContainer) to Units (UnitsContainer)
        self._dimensionality_cache = dict()

        #: Cache the unit name associated to user input. ('mV' -> 'millivolt')
        self._parse_unit_cache = dict()

        #: When performing a multiplication of units, interpret
        #: non-multiplicative units as their *delta* counterparts.
        self.default_as_delta = default_as_delta

        #: System name to be used by default.
        self._default_system_name = None

        # Determines if quantities with offset units are converted to their
        # base units on multiplication and division.
        self.autoconvert_offset_to_baseunit = autoconvert_offset_to_baseunit

        if filename == '':
            self.load_definitions('default_en.txt', True)
        elif filename is not None:
            self.load_definitions(filename)

        self.define(UnitDefinition('pi', 'π', (), ScaleConverter(math.pi)))

        self._build_cache()

        self.set_default_system(system)

    def __name__(self):
        return 'UnitRegistry'

    def __getattr__(self, item):
        return self.Unit(item)

    def __getitem__(self, item):
        logger.warning('Calling the getitem method from a UnitRegistry is deprecated. '
                       'use `parse_expression` method or use the registry as a callable.')
        return self.parse_expression(item)

    def __dir__(self):
        return list(self._units.keys()) + \
            ['define', 'load_definitions', 'get_name', 'get_symbol',
             'get_dimensionality', 'Quantity', 'wraps',
             'parse_units', 'parse_expression', 'pi_theorem',
             'convert', 'get_base_units']

    @property
    def default_format(self):
        """Default formatting string for quantities.
        """
        return self.Quantity.default_format

    @default_format.setter
    def default_format(self, value):
        self.Unit.default_format = value
        self.Quantity.default_format = value

    def get_group(self, name, create_if_needed=True):
        """Return a Group.

        :param name: Name of the group to be
        :param create_if_needed: Create a group if not Found. If False, raise an Exception.
        :return: Group
        """
        return self._gsmanager.get_group(name, create_if_needed)

    def get_system(self, name, create_if_needed=True):
        """Return a Group.

        :param registry:
        :param name: Name of the group to be
        :param create_if_needed: Create a group if not Found. If False, raise an Exception.
        :return: System
        """
        return self._gsmanager.get_system(name, create_if_needed)

    def set_default_system(self, name):
        # Test if exists
        if name:
            system = self._gsmanager.get_system(name, False)
            self._default_system_name = name
            self._base_units_cache = {}

    def add_context(self, context):
        """Add a context object to the registry.

        The context will be accessible by its name and aliases.

        Notice that this method will NOT enable the context. Use `enable_contexts`.
        """
        if context.name in self._contexts:
            logger.warning('The name %s was already registered for another context.',
                           context.name)
        self._contexts[context.name] = context
        for alias in context.aliases:
            if alias in self._contexts:
                logger.warning('The name %s was already registered for another context',
                               context.name)
            self._contexts[alias] = context

    def remove_context(self, name_or_alias):
        """Remove a context from the registry and return it.

        Notice that this methods will not disable the context. Use `disable_contexts`.
        """
        context = self._contexts[name_or_alias]

        del self._contexts[context.name]
        for alias in context.aliases:
            del self._contexts[alias]

        return context

    def enable_contexts(self, *names_or_contexts, **kwargs):
        """Enable contexts provided by name or by object.

        :param names_or_contexts: sequence of the contexts or contexts names/alias
        :param kwargs: keyword arguments for the context
        """

        # If present, copy the defaults from the containing contexts
        if self._active_ctx.defaults:
            kwargs = dict(self._active_ctx.defaults, **kwargs)

        # For each name, we first find the corresponding context
        ctxs = tuple((self._contexts[name] if isinstance(name, string_types) else name)
                     for name in names_or_contexts)

        # Check if the contexts have been checked first, if not we make sure
        # that dimensions are expressed in terms of base dimensions.
        for ctx in ctxs:
            if getattr(ctx, '_checked', False):
                continue
            for (src, dst), func in ctx.funcs.items():
                src_ = self._get_dimensionality(src)
                dst_ = self._get_dimensionality(dst)
                if src != src_ or dst != dst_:
                    ctx.remove_transformation(src, dst)
                    ctx.add_transformation(src_, dst_, func)
            ctx._checked = True

        # and create a new one with the new defaults.
        ctxs = tuple(Context.from_context(ctx, **kwargs)
                     for ctx in ctxs)

        # Finally we add them to the active context.
        self._active_ctx.insert_contexts(*ctxs)

    def disable_contexts(self, n=None):
        """Disable the last n enabled contexts.
        """
        if n is None:
            n = len(self._contexts)
        self._active_ctx.remove_contexts(n)

    @contextmanager
    def context(self, *names, **kwargs):
        """Used as a context manager, this function enables to activate a context
        which is removed after usage.

        :param names: name of the context.
        :param kwargs: keyword arguments for the contexts.

        Context are called by their name::

            >>> with ureg.context('one'):
            ...     pass

        If the context has an argument, you can specify its value as a keyword
        argument::

            >>> with ureg.context('one', n=1):
            ...     pass

        Multiple contexts can be entered in single call:

            >>> with ureg.context('one', 'two', n=1):
            ...     pass

        or nested allowing you to give different values to the same keyword argument::

            >>> with ureg.context('one', n=1):
            ...     with ureg.context('two', n=2):
            ...         pass

        A nested context inherits the defaults from the containing context::

            >>> with ureg.context('one', n=1):
            ...     with ureg.context('two'): # Here n takes the value of the upper context
            ...         pass

        """

        # Enable the contexts.
        self.enable_contexts(*names, **kwargs)

        try:
            # After adding the context and rebuilding the graph, the registry
            # is ready to use.
            yield self
        finally:
            # Upon leaving the with statement,
            # the added contexts are removed from the active one.
            self.disable_contexts(len(names))

    def define(self, definition):
        """Add unit to the registry.
        """
        if isinstance(definition, string_types):
            definition = Definition.from_string(definition)

        if isinstance(definition, DimensionDefinition):
            d, di = self._dimensions, None
        elif isinstance(definition, UnitDefinition):
            d, di = self._units, self._units_casei
            if definition.is_base:
                for dimension in definition.reference.keys():
                    if dimension in self._dimensions:
                        if dimension != '[]':
                            raise DefinitionSyntaxError('only one unit per dimension can be a base unit.')
                        continue

                    self.define(DimensionDefinition(dimension, '', (), None, is_base=True))

        elif isinstance(definition, PrefixDefinition):
            d, di = self._prefixes, None
        else:
            raise TypeError('{0} is not a valid definition.'.format(definition))

        def _adder(key, value, action=self._on_redefinition, selected_dict=d, casei_dict=di):
            if key in selected_dict:
                if action == 'raise':
                    raise RedefinitionError(key, type(value))
                elif action == 'warn':
                    logger.warning("Redefining '%s' (%s)", key, type(value))

            selected_dict[key] = value
            if casei_dict is not None:
                casei_dict[key.lower()].add(key)

        _adder(definition.name, definition)

        if definition.has_symbol:
            _adder(definition.symbol, definition)

        for alias in definition.aliases:
            if ' ' in alias:
                logger.warn('Alias cannot contain a space: ' + alias)

            _adder(alias, definition)

        # define additional "delta_" units for units with an offset
        if getattr(definition.converter, "offset", 0.0) != 0.0:
            d_name = 'delta_' + definition.name
            if definition.symbol:
                d_symbol = 'Δ' + definition.symbol
            else:
                d_symbol = None
            d_aliases = tuple('Δ' + alias for alias in definition.aliases)

            def prep(_name):
                if _name.startswith('['):
                    return '[delta_' + _name[1:]
                return 'delta_' + _name

            d_reference = UnitsContainer(dict((ref, value)
                                         for ref, value in definition.reference.items()))

            self.define(UnitDefinition(d_name, d_symbol, d_aliases,
                                       ScaleConverter(definition.converter.scale),
                                       d_reference, definition.is_base))

    def load_definitions(self, file, is_resource=False):
        """Add units and prefixes defined in a definition text file.
        """
        # Permit both filenames and line-iterables
        if isinstance(file, string_types):
            try:
                if is_resource:
                    with closing(pkg_resources.resource_stream(__name__, file)) as fp:
                        rbytes = fp.read()
                    return self.load_definitions(StringIO(rbytes.decode('utf-8')), is_resource)
                else:
                    with open(file, encoding='utf-8') as fp:
                        return self.load_definitions(fp, is_resource)
            except (RedefinitionError, DefinitionSyntaxError) as e:
                if e.filename is None:
                    e.filename = file
                raise e
            except Exception as e:
                msg = getattr(e, 'message', '') or str(e)
                raise ValueError('While opening {0}\n{1}'.format(file, msg))

        ifile = enumerate(file, 1)
        for no, line in ifile:
            line = line.strip()
            if not line or line.startswith('#'):
                continue
            if line.startswith('@import'):
                if is_resource:
                    path = line[7:].strip()
                else:
                    try:
                        path = os.path.dirname(file.name)
                    except AttributeError:
                        path = os.getcwd()
                    path = os.path.join(path, os.path.normpath(line[7:].strip()))
                self.load_definitions(path, is_resource)
            elif line.startswith('@context'):
                context = [line, ] + _capture_till_end(ifile)
                try:
                    self.add_context(Context.from_lines(context, self.get_dimensionality))
                except KeyError as e:
                    raise DefinitionSyntaxError('unknown dimension {0} in context'.format(str(e)), lineno=no)
            elif line.startswith('@system'):
                context = [line, ] + _capture_till_end(ifile)
                try:
                    self._gsmanager.add_system_from_lines(context, self.get_root_units)
                except KeyError as e:
                    raise DefinitionSyntaxError('unknown dimension {0} in context'.format(str(e)), lineno=no)
            elif line.startswith('@group'):
                context = [line, ] + _capture_till_end(ifile)
                try:
                    self._gsmanager.add_group_from_lines(context, self.define)
                except KeyError as e:
                    raise DefinitionSyntaxError('unknown dimension {0} in context'.format(str(e)), lineno=no)
            else:
                try:
                    self.define(Definition.from_string(line))
                except (RedefinitionError, DefinitionSyntaxError) as ex:
                    if ex.lineno is None:
                        ex.lineno = no
                    raise ex
                except Exception as ex:
                    logger.error("In line {0}, cannot add '{1}' {2}".format(no, line, ex))

    def _build_cache(self):
        """Build a cache of dimensionality and base units.
        """

        deps = dict((name, set(definition.reference.keys() if definition.reference else {}))
                    for name, definition in self._units.items())

        for unit_names in solve_dependencies(deps):
            for unit_name in unit_names:
                prefixed = False
                for p in self._prefixes.keys():
                    if p and unit_name.startswith(p):
                        prefixed = True
                        break
                if '[' in unit_name:
                    continue
                try:
                    uc = ParserHelper.from_word(unit_name)

                    bu = self._get_root_units(uc)
                    di = self._get_dimensionality(uc)

                    self._root_units_cache[uc] = bu
                    self._dimensionality_cache[uc] = di

                    if not prefixed:
                        if di not in self._dimensional_equivalents:
                            self._dimensional_equivalents[di] = set()

                        self._dimensional_equivalents[di].add(self._units[unit_name]._name)

                except Exception as e:
                    logger.warning('Could not resolve {0}: {1!r}'.format(unit_name, e))

    def get_name(self, name_or_alias, case_sensitive=True):
        """Return the canonical name of a unit.
        """

        if name_or_alias == 'dimensionless':
            return ''

        try:
            return self._units[name_or_alias]._name
        except KeyError:
            pass

        candidates = self._dedup_candidates(self.parse_unit_name(name_or_alias, case_sensitive))
        if not candidates:
            raise UndefinedUnitError(name_or_alias)
        elif len(candidates) == 1:
            prefix, unit_name, _ = candidates[0]
        else:
            logger.warning('Parsing {0} yield multiple results. '
                           'Options are: {1}'.format(name_or_alias, candidates))
            prefix, unit_name, _ = candidates[0]

        if prefix:
            name = prefix + unit_name
            symbol = self.get_symbol(name)
            prefix_def = self._prefixes[prefix]
            self._units[name] = UnitDefinition(name, symbol, (), prefix_def.converter,
                                               UnitsContainer({unit_name: 1}))
            return prefix + unit_name

        return unit_name

    def get_symbol(self, name_or_alias):
        """Return the preferred alias for a unit
        """
        candidates = self._dedup_candidates(self.parse_unit_name(name_or_alias))
        if not candidates:
            raise UndefinedUnitError(name_or_alias)
        elif len(candidates) == 1:
            prefix, unit_name, _ = candidates[0]
        else:
            logger.warning('Parsing {0} yield multiple results. '
                           'Options are: {1!r}'.format(name_or_alias, candidates))
            prefix, unit_name, _ = candidates[0]

        return self._prefixes[prefix].symbol + self._units[unit_name].symbol

    def _get_symbol(self, name):
        return self._units[name].symbol

    def get_dimensionality(self, input_units):
        """Convert unit or dict of units or dimensions to a dict of base dimensions
        dimensions

        :param input_units:
        :return: dimensionality
        """
        input_units = to_units_container(input_units)

        return self._get_dimensionality(input_units)

    def _get_dimensionality(self, input_units):
        """ Convert a UnitsContainer to base dimensions.

        :param input_units:
        :return: dimensionality
        """
        if not input_units:
            return UnitsContainer()

        if input_units in self._dimensionality_cache:
            return self._dimensionality_cache[input_units]

        accumulator = defaultdict(float)
        self._get_dimensionality_recurse(input_units, 1.0, accumulator)

        if '[]' in accumulator:
            del accumulator['[]']

        dims = UnitsContainer(dict((k, v) for k, v in accumulator.items()
                                   if v != 0.0))

        self._dimensionality_cache[input_units] = dims

        return dims

    def _get_dimensionality_recurse(self, ref, exp, accumulator):
        for key in ref:
            exp2 = exp*ref[key]
            if _is_dim(key):
                reg = self._dimensions[key]
                if reg.is_base:
                    accumulator[key] += exp2
                elif reg.reference is not None:
                    self._get_dimensionality_recurse(reg.reference, exp2, accumulator)
            else:
                reg = self._units[self.get_name(key)]
                if reg.reference is not None:
                    self._get_dimensionality_recurse(reg.reference, exp2, accumulator)

    def get_root_units(self, input_units, check_nonmult=True):
        """Convert unit or dict of units to the root units.

        If any unit is non multiplicative and check_converter is True,
        then None is returned as the multiplicative factor.

        :param input_units: units
        :type input_units: UnitsContainer or str
        :param check_nonmult: if True, None will be returned as the
                              multiplicative factor if a non-multiplicative
                              units is found in the final Units.
        :return: multiplicative factor, base units
        """
        input_units = to_units_container(input_units)

        f, units = self._get_root_units(input_units, check_nonmult)

        return f, self.Unit(units)

    def _get_root_units(self, input_units, check_nonmult=True):
        """Convert unit or dict of units to the root units.

        If any unit is non multiplicative and check_converter is True,
        then None is returned as the multiplicative factor.

        :param input_units: units
        :type input_units: UnitsContainer or dict
        :param check_nonmult: if True, None will be returned as the
                              multiplicative factor if a non-multiplicative
                              units is found in the final Units.
        :return: multiplicative factor, base units
        """
        if not input_units:
            return 1., UnitsContainer()

        # The cache is only done for check_nonmult=True
        if check_nonmult and input_units in self._root_units_cache:
            return self._root_units_cache[input_units]

        accumulators = [1., defaultdict(float)]
        self._get_root_units_recurse(input_units, 1.0, accumulators)

        factor = accumulators[0]
        units = UnitsContainer(dict((k, v) for k, v in accumulators[1].items()
                                    if v != 0.))

        # Check if any of the final units is non multiplicative and return None instead.
        if check_nonmult:
            for unit in units.keys():
                if not self._units[unit].converter.is_multiplicative:
                    return None, units

        if check_nonmult:
            self._root_units_cache[input_units] = factor, units

        return factor, units

    def get_base_units(self, input_units, check_nonmult=True, system=None):
        """Convert unit or dict of units to the base units.

        If any unit is non multiplicative and check_converter is True,
        then None is returned as the multiplicative factor.

        :param input_units: units
        :type input_units: UnitsContainer or str
        :param check_nonmult: if True, None will be returned as the
                              multiplicative factor if a non-multiplicative
                              units is found in the final Units.
        :return: multiplicative factor, base units
        """
        input_units = to_units_container(input_units)

        f, units = self._get_base_units(input_units, check_nonmult, system)

        return f, self.Unit(units)

    def _get_base_units(self, input_units, check_nonmult=True, system=None):
        """
        :param registry:
        :param input_units:
        :param check_nonmult:
        :param system: System
        :return:
        """

        # The cache is only done for check_nonmult=True
        if check_nonmult and input_units in self._base_units_cache:
            return self._base_units_cache[input_units]

        factor, units = self.get_root_units(input_units, check_nonmult)

        if system is None:
            system = self._default_system_name

        if not system:
            return factor, units

        # This will not be necessary after integration with the registry as it has a UnitsContainer intermediate
        units = to_units_container(units, self)

        destination_units = UnitsContainer()

        bu = self._gsmanager.get_system(system, False).base_units

        for unit, value in units.items():
            if unit in bu:
                new_unit = bu[unit]
                new_unit = to_units_container(new_unit, self)
                destination_units *= new_unit ** value
            else:
                destination_units *= UnitsContainer({unit: value})

        base_factor = self.convert(factor, units, destination_units)

        if check_nonmult:
            self._base_units_cache[input_units] = base_factor, destination_units

        return base_factor, destination_units

    def _get_root_units_recurse(self, ref, exp, accumulators):
        for key in ref:
            exp2 = exp*ref[key]
            key = self.get_name(key)
            reg = self._units[key]
            if reg.is_base:
                accumulators[1][key] += exp2
            else:
                accumulators[0] *= reg._converter.scale ** exp2
                if reg.reference is not None:
                    self._get_root_units_recurse(reg.reference, exp2,
                                                 accumulators)

    def get_compatible_units(self, input_units, group_or_system=None):
        """
        """
        input_units = to_units_container(input_units)

        equiv = self._get_compatible_units(input_units, group_or_system)

        return frozenset(self.Unit(eq) for eq in equiv)

    def _get_compatible_units(self, input_units, group_or_system=None):
        """
        """
        if not input_units:
            return frozenset()

        src_dim = self._get_dimensionality(input_units)

        ret = self._dimensional_equivalents[src_dim]

        if self._active_ctx:
            nodes = find_connected_nodes(self._active_ctx.graph, src_dim)
            ret = set()
            if nodes:
                for node in nodes:
                    ret |= self._dimensional_equivalents[node]

        if group_or_system:
            members = self._gsmanager[group_or_system].members
            return frozenset(ret.intersection(members))

        return ret

    def convert(self, value, src, dst, inplace=False):
        """Convert value from some source to destination units.

        :param value: value
        :param src: source units.
        :type src: Quantity or str
        :param dst: destination units.
        :type dst: Quantity or str

        :return: converted value
        """
        src = to_units_container(src, self)

        dst = to_units_container(dst, self)

        return self._convert(value, src, dst, inplace)

    def _convert(self, value, src, dst, inplace=False):
        """Convert value from some source to destination units.

        :param value: value
        :param src: source units.
        :type src: UnitsContainer
        :param dst: destination units.
        :type dst: UnitsContainer

        :return: converted value
        """
        if src == dst:
            return value

        src_dim = self._get_dimensionality(src)
        dst_dim = self._get_dimensionality(dst)

        # If there is an active context, we look for a path connecting source and
        # destination dimensionality. If it exists, we transform the source value
        # by applying sequentially each transformation of the path.
        if self._active_ctx:
            path = find_shortest_path(self._active_ctx.graph, src_dim, dst_dim)
            if path:
                src = self.Quantity(value, src)
                for a, b in zip(path[:-1], path[1:]):
                    src = self._active_ctx.transform(a, b, self, src)

                value, src = src._magnitude, src._units

                src_dim = self._get_dimensionality(src)

        # If the source and destination dimensionality are different,
        # then the conversion cannot be performed.

        if src_dim != dst_dim:
            raise DimensionalityError(src, dst, src_dim, dst_dim)

        # Conversion needs to consider if non-multiplicative (AKA offset
        # units) are involved. Conversion is only possible if src and dst
        # have at most one offset unit per dimension.
        src_offset_units = [(u, e) for u, e in src.items()
                            if not self._units[u].is_multiplicative]
        dst_offset_units = [(u, e) for u, e in dst.items()
                            if not self._units[u].is_multiplicative]

        # For offset units we need to check if the conversion is allowed.
        if src_offset_units or dst_offset_units:

            # Validate that not more than one offset unit is present
            if len(src_offset_units) > 1 or len(dst_offset_units) > 1:
                raise DimensionalityError(
                    src, dst, src_dim, dst_dim,
                    extra_msg=' - more than one offset unit.')

            # validate that offset unit is not used in multiplicative context
            if ((len(src_offset_units) == 1 and len(src) > 1)
                    or (len(dst_offset_units) == 1 and len(dst) > 1)
                    and not self.autoconvert_offset_to_baseunit):
                raise DimensionalityError(
                    src, dst, src_dim, dst_dim,
                    extra_msg=' - offset unit used in multiplicative context.')

            # Validate that order of offset unit is exactly one.
            if src_offset_units:
                if src_offset_units[0][1] != 1:
                    raise DimensionalityError(
                        src, dst, src_dim, dst_dim,
                        extra_msg=' - offset units in higher order.')
            else:
                if dst_offset_units[0][1] != 1:
                    raise DimensionalityError(
                        src, dst, src_dim, dst_dim,
                        extra_msg=' - offset units in higher order.')

        # Here we convert only the offset quantities. Any remaining scaled
        # quantities will be converted later.

        # TODO: Shouldn't this (until factor, units) be inside the If above?

        # clean src from offset units by converting to reference
        for u, e in src_offset_units:
            value = self._units[u].converter.to_reference(value, inplace)
        src = src.remove([u for u, e in src_offset_units])

        # clean dst units from offset units
        dst = dst.remove([u for u, e in dst_offset_units])

        # Here src and dst have only multiplicative units left. Thus we can
        # convert with a factor.
        factor, units = self._get_root_units(src / dst)

        # factor is type float and if our magnitude is type Decimal then
        # must first convert to Decimal before we can '*' the values
        if isinstance(value, Decimal):
            factor = Decimal(str(factor))

        if inplace:
            value *= factor
        else:
            value = value * factor

        # Finally convert to offset units specified in destination
        for u, e in dst_offset_units:
            value = self._units[u].converter.from_reference(value, inplace)

        return value

    def pi_theorem(self, quantities):
        """Builds dimensionless quantities using the Buckingham π theorem

        :param quantities: mapping between variable name and units
        :type quantities: dict
        :return: a list of dimensionless quantities expressed as dicts
        """
        return pi_theorem(quantities, self)

    def _dedup_candidates(self, candidates):
        """Given a list of units, remove those with different names but equal value.
        """
        candidates = tuple(candidates)
        if len(candidates) < 2:
            return candidates

        unique = [candidates[0]]
        for c in candidates[2:]:
            for u in unique:
                if c == u:
                    break
            else:
                unique.append(c)

        return tuple(unique)

    def parse_unit_name(self, unit_name, case_sensitive=True):
        """Parse a unit to identify prefix, unit name and suffix
        by walking the list of prefix and suffix.
        """
        stw = unit_name.startswith
        edw = unit_name.endswith
        for suffix, prefix in itertools.product(self._suffixes, self._prefixes):
            if stw(prefix) and edw(suffix):
                name = unit_name[len(prefix):]
                if suffix:
                    name = name[:-len(suffix)]
                    if len(name) == 1:
                        continue
                if case_sensitive:
                    if name in self._units:
                        yield (self._prefixes[prefix]._name,
                               self._units[name]._name,
                               self._suffixes[suffix])
                else:
                    for real_name in self._units_casei.get(name.lower(), ()):
                        yield (self._prefixes[prefix]._name,
                               self._units[real_name]._name,
                               self._suffixes[suffix])

    def parse_units(self, input_string, as_delta=None):
        """Parse a units expression and returns a UnitContainer with
        the canonical names.

        The expression can only contain products, ratios and powers of units.

        :param as_delta: if the expression has multiple units, the parser will
                         interpret non multiplicative units as their `delta_` counterparts.

        :raises:
            :class:`pint.UndefinedUnitError` if a unit is not in the registry
            :class:`ValueError` if the expression is invalid.
        """
        units = self._parse_units(input_string, as_delta)
        return self.Unit(units)

    def _parse_units(self, input_string, as_delta=None):
        """
        """
        if input_string in self._parse_unit_cache:
            return self._parse_unit_cache[input_string]

        if as_delta is None:
            as_delta = self.default_as_delta

        if not input_string:
            return UnitsContainer()

        units = ParserHelper.from_string(input_string)
        if units.scale != 1:
            raise ValueError('Unit expression cannot have a scaling factor.')

        ret = {}
        many = len(units) > 1
        for name in units:
            cname = self.get_name(name)
            value = units[name]
            if not cname:
                continue
            if as_delta and (many or (not many and value != 1)):
                definition = self._units[cname]
                if not definition.is_multiplicative:
                    cname = 'delta_' + cname
            ret[cname] = value

        ret = UnitsContainer(ret)

        self._parse_unit_cache[input_string] = ret

        return ret
    
    def _eval_token(self, token, case_sensitive=True, **values):
        token_type = token[0]
        token_text = token[1]
        if token_type == NAME:
            if token_text == 'pi':
                return self.Quantity(math.pi)
            elif token_text in values:
                return self.Quantity(values[token_text])
            else:
                return self.Quantity(1, UnitsContainer({self.get_name(token_text, 
                                                                      case_sensitive=case_sensitive) : 1}))
        elif token_type == NUMBER:
            return ParserHelper.eval_token(token)
        else:
            raise Exception('unknown token type')

    def parse_expression(self, input_string, case_sensitive=True, **values):
        """Parse a mathematical expression including units and return a quantity object.

        Numerical constants can be specified as keyword arguments and will take precedence
        over the names defined in the registry.
        """
        
        if not input_string:
            return self.Quantity(1)

        input_string = string_preprocessor(input_string)
        gen = tokenizer(input_string)
        
        return build_eval_tree(gen).evaluate(lambda x : self._eval_token(x, case_sensitive=case_sensitive,
                                                                          **values))

    __call__ = parse_expression

    def wraps(self, ret, args, strict=True):
        """Wraps a function to become pint-aware.

        Use it when a function requires a numerical value but in some specific
        units. The wrapper function will take a pint quantity, convert to the units
        specified in `args` and then call the wrapped function with the resulting
        magnitude.

        The value returned by the wrapped function will be converted to the units
        specified in `ret`.

        Use None to skip argument conversion.
        Set strict to False, to accept also numerical values.

        :param ret: output units.
        :param args: iterable of input units.
        :param strict: boolean to indicate that only quantities are accepted.
        :return: the wrapped function.
        :raises:
            :class:`ValueError` if strict and one of the arguments is not a Quantity.
        """

        Q_ = self.Quantity

        if not isinstance(args, (list, tuple)):
            args = (args, )

        units = [to_units_container(arg, self) for arg in args]

        if isinstance(ret, (list, tuple)):
            ret = ret.__class__([to_units_container(arg, self) for arg in ret])
        elif isinstance(ret, string_types):
            ret = self.parse_units(ret)

        def decorator(func):
            assigned = tuple(attr for attr in functools.WRAPPER_ASSIGNMENTS if hasattr(func, attr))
            updated = tuple(attr for attr in functools.WRAPPER_UPDATES if hasattr(func, attr))
            @functools.wraps(func, assigned=assigned, updated=updated)
            def wrapper(*values, **kw):
                new_args = []
                for unit, value in zip(units, values):
                    if unit is None:
                        new_args.append(value)
                    elif isinstance(value, Q_):
                        new_args.append(self._convert(value._magnitude,
                                                      value._units, unit))
                    elif not strict:
                        new_args.append(value)
                    else:
                        raise ValueError('A wrapped function using strict=True requires '
                                         'quantity for all arguments with not None units. '
                                         '(error found for {0}, {1})'.format(unit, value))

                result = func(*new_args, **kw)

                if isinstance(ret, (list, tuple)):
                    return ret.__class__(res if unit is None else Q_(res, unit)
                                         for unit, res in zip(ret, result))
                elif ret is not None:
                    return Q_(result, ret)

                return result
            return wrapper
        return decorator

    def check(self, *args):
        """Decorator to for quantity type checking for function inputs.

        Use it to ensure that the decorated function input parameters match
        the expected type of pint quantity.

        Use None to skip argument checking.

        :param args: iterable of input units.
        :return: the wrapped function.
        :raises:
            :class:`DimensionalityError` if the parameters don't match dimensions
        """
        dimensions = [self.get_dimensionality(dim) for dim in args]

        def decorator(func):
            assigned = tuple(attr for attr in functools.WRAPPER_ASSIGNMENTS if hasattr(func, attr))
            updated = tuple(attr for attr in functools.WRAPPER_UPDATES if hasattr(func, attr))

            @functools.wraps(func, assigned=assigned, updated=updated)
            def wrapper(*values, **kwargs):
                for dim, value in zip_longest(dimensions, values):
                    if dim and value.dimensionality != dim:
                        raise DimensionalityError(value, 'a quantity of',
                                                  value.dimensionality, dim)
                return func(*values, **kwargs)
            return wrapper
        return decorator


def build_unit_class(registry):

    class Unit(_Unit):
        pass

    Unit._REGISTRY = registry
    return Unit


def build_quantity_class(registry, force_ndarray=False):
    from .quantity import _Quantity

    class Quantity(_Quantity):
        pass

    Quantity._REGISTRY = registry
    Quantity.force_ndarray = force_ndarray

    return Quantity


def build_measurement_class(registry, force_ndarray=False):
    from .measurement import _Measurement, ufloat

    if ufloat is None:
        class Measurement(object):

            def __init__(self, *args):
                raise RuntimeError("Pint requires the 'uncertainties' package to create a Measurement object.")

    else:
        class Measurement(_Measurement, registry.Quantity):
            pass

    Measurement._REGISTRY = registry
    Measurement.force_ndarray = force_ndarray

    return Measurement


class LazyRegistry(object):

    def __init__(self, args=None, kwargs=None):
        self.__dict__['params'] = args or (), kwargs or {}

    def __init(self):
        args, kwargs = self.__dict__['params']
        kwargs['on_redefinition'] = 'raise'
        self.__class__ = UnitRegistry
        self.__init__(*args, **kwargs)

    def __getattr__(self, item):
        if item == '_on_redefinition':
            return 'raise'
        self.__init()
        return getattr(self, item)

    def __setattr__(self, key, value):
        if key == '__class__':
            super(LazyRegistry, self).__setattr__(key, value)
        else:
            self.__init()
            setattr(self, key, value)

    def __getitem__(self, item):
        self.__init()
        return self[item]

    def __call__(self, *args, **kwargs):
        self.__init()
        return self(*args, **kwargs)



#!/usr/bin/env python

#############################################################################
##
# This file is part of Taurus
##
# http://taurus-scada.org
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Taurus is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Taurus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
#############################################################################

"""
taurustrend.py: Generic trend widget for Taurus
"""
__all__ = ["TaurusTrend2DDialog"]

from guiqwt.plot import ImageDialog
from taurus.external.qt import Qt
import taurus.core
from taurus.qt.qtgui.base import TaurusBaseWidget
from taurus.qt.qtgui.extra_guiqwt.image import TaurusTrend2DItem
from taurus.qt.qtgui.extra_guiqwt.tools import (TaurusModelChooserTool,
                                                TimeAxisTool, AutoScrollTool)


class TaurusTrend2DDialog(ImageDialog, TaurusBaseWidget):
    """
    This is a widget for displaying trends from 1D Taurus attributes (i.e.,
    representing the variation over time of a 1D array). Sometimes this kind of
    plots are also known as "spectrograms".

    The widget shows a 3D plot (Z represented with colors) where the values in
    the 1D array are plotted in the Y-Z plane and are stacked along the X axis.
    """
    _modifiableByUser = True

    def __init__(self, parent=None, designMode=False, toolbar=True,
                 stackMode='deltatime', buffersize=512, options=None, **kwargs):
        """see :class:`guiqwt.plot.ImageDialog` for other valid initialization
        parameters"""
        defaultOptions = dict(lock_aspect_ratio=False)
        if options is not None:
            defaultOptions.update(options)
        ImageDialog.__init__(self, parent=parent, toolbar=toolbar,
                             options=defaultOptions, **kwargs)
        TaurusBaseWidget.__init__(self, "TaurusTrend2DDialog")
        self.trendItem = None
        self.buffersize = buffersize
        self._useArchiving = False
        self._stackMode = stackMode
        self.setStackMode(stackMode)
        self.setWindowFlags(Qt.Qt.Widget)
        # add some tools
        for toolklass in (TaurusModelChooserTool, AutoScrollTool):
            self.add_tool(toolklass)
        self.get_tool(TaurusModelChooserTool).singleModel = True
        self.setModifiableByUser(self._modifiableByUser)
        self.setContextMenuPolicy(Qt.Qt.CustomContextMenu)
        self.registerConfigDelegate(self.get_tool(AutoScrollTool))

    def keyPressEvent(self, event):
        if(event.key() == Qt.Qt.Key_Escape):
            event.ignore()
        else:
            ImageDialog.keyPressEvent(self, event)

    def setStackMode(self, mode):
        """set the type of stack to be used. This determines how X values are
        interpreted:

            - as timestamps ('datetime')
            - as time deltas ('timedelta')
            - as event numbers ('event')

        :param mode:(one of 'datetime', 'timedelta' or 'event')
        """
        mode = str(mode)
        if mode == 'datetime':
            self.add_tool(TimeAxisTool)
            timetool = self.get_tool(TimeAxisTool)
            timetool.set_scale_y_t(True)
        elif mode == 'deltatime':
            from taurus.qt.qtgui.plot import DeltaTimeScaleEngine
            plot = self.get_plot()
            DeltaTimeScaleEngine.enableInAxis(plot, plot.xBottom, rotation=-45)
        elif mode == 'event':
            plot = self.get_plot()
            scaleEngine = plot.axisScaleEngine(plot.xBottom)
            if hasattr(scaleEngine, 'disableInAxis'):
                scaleEngine.disableInAxis(plot, plot.xBottom)
        else:
            self.error('Unknown stack mode "%s"' % repr(mode))
            return
        self._stackMode = mode
        if hasattr(self.trendItem, 'stackMode'):
            self.trendItem.stackMode = mode

    def getStackMode(self):
        return self._stackMode

    def resetStackMode(self):
        self.setStackMode('datetime')

    def getModelClass(self):
        """reimplemented from :class:`TaurusBaseWidget`"""
        return taurus.core.taurusattribute.TaurusAttribute

    def setModel(self, model):
        """reimplemented from :class:`TaurusBaseWidget`"""
        plot = self.get_plot()
        if self.trendItem is not None:
            plot.del_item(self.trendItem)
        self.trendItem = TaurusTrend2DItem(
            stackMode=self.getStackMode(), buffersize=self.buffersize)
        self.trendItem.setModel(model)
        plot.add_item(self.trendItem)
        self.trendItem.set_readonly(not self.isModifiableByUser())
        plot.set_axis_title(plot.colormap_axis, 'value')
        plot.set_axis_unit('left', 'index')
        try:
            plot.set_axis_title(
                'left', self.trendItem.getModelObj().getSimpleName())
        except:
            self.debug('cannot set title for left axis')
            self.traceback()
        try:
            unit = self.trendItem.getModelObj().getConfig().getUnit() or ''
            plot.set_axis_unit(plot.colormap_axis, unit)
        except:
            self.debug('cannot set units for colormap axis')
            self.traceback()

        self.trendItem.dataChanged.connect(self.update_cross_sections)

    def getModel(self):
        """reimplemented from :class:`TaurusBaseWidget`"""
        if self.trendItem is None:
            return None
        else:
            return self.trendItem.getModel()

    def setUseArchiving(self, enable):
        """enables/disables looking up in the archiver for data stored before
        the Trend was started

        :param enable: (bool) if True, archiving values will be used if
                       available
        """
        if not self._stackMode == 'datetime':
            self.info('ignoring setUseArchiving. Reason: not in X time scale')
        self._useArchiving = enable

    def getUseArchiving(self):
        """whether TaurusTrend is looking for data in the archiver when needed

        :return: (bool)

        .. seealso:: :meth:`setUseArchiving`
        """
        return self._useArchiving

    def resetUseArchiving(self):
        """Same as setUseArchiving(False)"""
        self.setUseArchiving(False)

    def setMaxDataBufferSize(self, maxSize):
        """sets the maximum number of events that will be stacked

        :param maxSize: (int) the maximum limit

        .. seealso:: :class:`TaurusTrendSet`
        """
        if self.trendItem is not None:
            self.trendItem.setBufferSize(maxSize)

        self.buffersize = maxSize

    def getMaxDataBufferSize(self):
        """returns the maximum number of events that can be plotted in the trend

        :return: (int)
        """
        return self.buffersize

    def resetMaxDataBufferSize(self):
        """Same as setMaxDataBufferSize(512)  (i.e. 512 events)"""
        self.setMaxDataBufferSize(512)

    @classmethod
    def getQtDesignerPluginInfo(cls):
        """reimplemented from :class:`TaurusBaseWidget`"""
        ret = TaurusBaseWidget.getQtDesignerPluginInfo()
        ret['module'] = 'taurus.qt.qtgui.plot'
        ret['group'] = 'Taurus Display'
        ret['icon'] = 'designer:qwtplot.png'
        return ret

    def setModifiableByUser(self, modifiable):
        """reimplemented from :class:`TaurusBaseWidget`"""
        self.get_tool(TaurusModelChooserTool).action.setEnabled(modifiable)
        self.get_plot().set_items_readonly(not modifiable)
        TaurusBaseWidget.setModifiableByUser(self, modifiable)

    model = Qt.pyqtProperty("QString", getModel, setModel,
                            TaurusBaseWidget.resetModel)
    # @todo uncomment this when archiving is supported
    useArchiving = Qt.pyqtProperty("bool", getUseArchiving, setUseArchiving,
                                   resetUseArchiving)
    maxDataBufferSize = Qt.pyqtProperty("int", getMaxDataBufferSize,
                                        setMaxDataBufferSize,
                                        resetMaxDataBufferSize)
    stackMode = Qt.pyqtProperty("QString", getStackMode, setStackMode,
                                resetStackMode)
    modifiableByUser = Qt.pyqtProperty("bool",
                                       TaurusBaseWidget.isModifiableByUser,
                                       setModifiableByUser,
                                       TaurusBaseWidget.resetModifiableByUser)


def taurusTrend2DMain():
    from taurus.qt.qtgui.application import TaurusApplication
    import taurus.core
    import sys

    # prepare options
    parser = taurus.core.util.argparse.get_taurus_parser()
    parser.set_usage("%prog [options] <model>")
    parser.set_description('a Taurus application for plotting trends of ' +
                           'arrays (aka "spectrograms")')
    parser.add_option("-x", "--x-axis-mode", dest="x_axis_mode", default='d',
                      metavar="t|d|e",
                      help=("interpret X values as timestamps (t), " +
                            "time deltas (d) or event numbers (e). " +
                            "Accepted values: t|d|e")
                      )
    parser.add_option("-b", "--buffer", dest="max_buffer_size", default='512',
                      help=("maximum number of values to be stacked " +
                            "(when reached, the oldest values will be " +
                            "discarded)")
                      )
    parser.add_option("-a", "--use-archiving",
                      action="store_true", dest="use_archiving", default=False)
    parser.add_option("--demo", action="store_true", dest="demo",
                      default=False, help="show a demo of the widget")
    parser.add_option("--window-name", dest="window_name",
                      default="Taurus Trend 2D", help="Name of the window")

    app = TaurusApplication(cmd_line_parser=parser, app_name="Taurus Trend 2D",
                            app_version=taurus.Release.version)
    args = app.get_command_line_args()
    options = app.get_command_line_options()

    # check & process options
    stackModeMap = dict(t='datetime', d='deltatime', e='event')
    if options.x_axis_mode.lower() not in stackModeMap:
        parser.print_help(sys.stderr)
        sys.exit(1)

    stackMode = stackModeMap[options.x_axis_mode.lower()]

    if options.demo:
        args.append('eval:x=linspace(0,3,40);t=rand();sin(x+t)')

    w = TaurusTrend2DDialog(stackMode=stackMode, wintitle=options.window_name,
                            buffersize=int(options.max_buffer_size))

    # set archiving
    if options.use_archiving:
        raise NotImplementedError('Archiving support is not yet implemented')
        w.setUseArchiving(True)

    # set model
    if len(args) == 1:
        w.setModel(args[0])
    else:
        parser.print_help(sys.stderr)
        sys.exit(1)

    w.show()
    sys.exit(app.exec_())


if __name__ == "__main__":
    taurusTrend2DMain()



#!/usr/bin/env python

#############################################################################
##
# This file is part of Taurus
##
# http://taurus-scada.org
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Taurus is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Taurus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
###########################################################################

"""This configuration contains base modules and classes that may be used
by specific TaurusGui-based GUIs"""

__docformat__ = 'restructuredtext'

import os
import sys
from lxml import etree
from taurus.qt.qtgui.util import ExternalAppAction
from taurus.qt.qtgui.util import TaurusWidgetFactory
from taurus.core.util.log import Logger

# this is here only for backwards compatibility. It should not be used at all


class Qt_Qt:
    LeftDockWidgetArea = 1
    RightDockWidgetArea = 2
    BottomDockWidgetArea = 3
    TopDockWidgetArea = 4

TAURUSGUI_AREAS = {'Left': Qt_Qt.LeftDockWidgetArea,
                   'Right': Qt_Qt.RightDockWidgetArea,
                   'Top': Qt_Qt.TopDockWidgetArea,
                   'Bottom': Qt_Qt.BottomDockWidgetArea}


class ExternalApp(object):
    '''
    A description of an external application.
    Uses the same initialization as that of :class:`ExternalAppAction`
    Use :meth:`getAction` to obtain an instance of a :class:`ExternalAppAction`
    '''

    def __init__(self, *args, **kwargs):
        ''' see :meth:`ExternalAppAction.__init__`'''
        self.args = args
        self.kwargs = kwargs

    def getAction(self):
        '''
        Returns a :class:`ExternalAppAction` with the values used when
        initializing this ExternalApp instance

        :return: (ExternalAppAction)
        '''
        return ExternalAppAction(*self.args, **self.kwargs)

    @staticmethod
    def fromXml(xmlstring):
        ''' returns a ExternalApp object based on the xml string provided

        :param xmlstring: (unicode) XML code defining the values for the
                          cmdargs, text, icon and parent variables

        :return: (ExternalApp) an instance of ExternalApp
        '''
        try:
            root = etree.fromstring(xmlstring)
        except:
            raise ValueError('Invalid XML syntax')

        commandNode = root.find("command")
        if (commandNode is not None) and (commandNode.text is not None):
            command = commandNode.text
        else:
            raise ValueError('Invalid XML: <command> is mandatory')

        paramsNode = root.find("params")
        if (paramsNode is not None) and (paramsNode.text is not None):
            params = paramsNode.text
        else:
            params = ''

        textNode = root.find("text")
        if (textNode is not None) and (textNode.text is not None):
            text = textNode.text
        else:
            text = None

        iconNode = root.find("icon")
        if (iconNode is not None) and (iconNode.text is not None):
            icon = iconNode.text
        else:
            icon = None

        return ExternalApp(" ".join((command, params)), text=text, icon=icon)


class TaurusGuiComponentDescription(object):
    '''
    A base class for describing a taurusgui component.
    '''

    def __init__(self, name, classname=None, modulename=None, widgetname=None,
                 sharedDataWrite=None, sharedDataRead=None,
                 model=None, floating=True, **kwargs):
        self._name = name
        self._modulename = modulename
        self.setClassname(classname)
        self.setWidgetname(widgetname)
        if self.classname is None and (self.modulename is None or self.widgetname is None):
            raise ValueError(
                'Module info must be given (except if passing a Taurus class name)')
        self._floating = floating
        if sharedDataWrite is None:
            sharedDataWrite = {}
        self._sharedDataWrite = sharedDataWrite
        if sharedDataRead is None:
            sharedDataRead = {}
        self._sharedDataRead = sharedDataRead
        self._model = model

    def getName(self):
        return self._name

    def setName(self, name):
        self._name = name

    def getClassname(self):
        return self._classname

    def setClassname(self, classname):
        if classname is not None and '.' in classname:
            modulename, classname = classname.rsplit('.', 1)
            self.setModulename(modulename)
        self._classname = classname

    def getModulename(self):
        return self._modulename

    def setModulename(self, modulename):
        self._modulename = modulename

    def getWidgetname(self):
        return self._widgetname

    def setWidgetname(self, widgetname):
        if widgetname is not None and '.' in widgetname:
            modulename, widgetname = widgetname.rsplit('.', 1)
            self.setModulename(modulename)
        self._widgetname = widgetname

    def getArea(self):
        raise DeprecationWarning('getArea is deprecated')
        return self._area

    def setArea(self, area):
        raise DeprecationWarning('setArea is deprecated')
        self._area = area

    def isFloating(self):
        return self._floating

    def setFloating(self, floating):
        self._floating = floating

    def getSharedDataWrite(self):
        return self._sharedDataWrite

    def setSharedDataWrite(self, sharedDataWrite):
        self._sharedDataWrite = sharedDataWrite

    def getSharedDataRead(self):
        return self._sharedDataRead

    def setSharedDataRead(self, sharedDataRead):
        self._sharedDataRead = sharedDataRead

    def getModel(self):
        return self._model

    def setModel(self, model):
        self._model = model

    def getWidget(self, sdm=None, setModel=True):
        ''' Returns the widget to be inserted in the panel

        :param sdm: (SharedDataManager) if given, the widget will be registered as reader
                    and/or writer in this manager as defined by the sharedDataRead and sharedDataWrite properties
        :param setModel: (bool) if True (default) the widget will be given the model deined in the model property

        :return: (QWidget) a new widget instance matching the description
        '''
        # instantiate the widget
        if self.modulename is None:
            klass = TaurusWidgetFactory().getWidgetClass(self.classname)
            w = klass()
        else:
            module = __import__(self.modulename, fromlist=[''])
            if self.classname is None:
                w = getattr(module, self.widgetname)
            else:
                klass = getattr(module, self.classname)
                w = klass()
        # set the model if setModel is True
        if self.model is not None and setModel:
            w.setModel(self.model)
        # connect (if an sdm is given)
        if sdm is not None:
            for dataUID, signalname in self.sharedDataWrite.iteritems():
                sdm.connectWriter(dataUID, w, signalname)
            for dataUID, slotname in self.sharedDataRead.iteritems():
                sdm.connectReader(dataUID, getattr(w, slotname))
        # set the name
        w.name = self.name
        return w

    def toXml(self):
        '''Returns a (unicode) XML code defining the PanelDescription object

        :return: xmlstring
        '''

        root = etree.Element("PanelDescription")
        name = etree.SubElement(root, "name")
        name.text = self._name
        classname = etree.SubElement(root, "classname")
        classname.text = self._classname
        modulename = etree.SubElement(root, "modulename")
        modulename.text = self._modulename
        widgetname = etree.SubElement(root, "widgetname")
        widgetname.text = self._widgetname
        floating = etree.SubElement(root, "floating")
        floating.text = str(self._floating)

        sharedDataWrite = etree.SubElement(root, "sharedDataWrite")
        for k, v in self._sharedDataWrite.iteritems():
            item = etree.SubElement(
                sharedDataWrite, "item", datauid=k, signalName=v)

        sharedDataRead = etree.SubElement(root, "sharedDataRead")
        for k, v in self._sharedDataRead.iteritems():
            item = etree.SubElement(
                sharedDataRead, "item", datauid=k, slotName=v)

        model = etree.SubElement(root, "model")
        model.text = self._model

        return etree.tostring(root, pretty_print=True)

    @staticmethod
    def fromXml(xmlstring):
        '''returns a PanelDescription object based on the xml string provided

        :param xmlstring: (unicode) XML code defining the values for the args
                          needed to initialize PanelDescription.

        :return: (PanelDescription) object
        '''

        try:
            root = etree.fromstring(xmlstring)
        except:
            return None

        nameNode = root.find("name")
        if (nameNode is not None) and (nameNode.text is not None):
            name = nameNode.text
        else:
            return None

        classnameNode = root.find("classname")
        if (classnameNode is not None) and (classnameNode.text is not None):
            classname = classnameNode.text
        else:
            classname = None

        modulenameNode = root.find("modulename")
        if (modulenameNode is not None) and (modulenameNode.text is not None):
            modulename = modulenameNode.text
        else:
            modulename = None

        widgetnameNode = root.find("widgetname")
        if (widgetnameNode is not None) and (widgetnameNode.text is not None):
            widgetname = widgetnameNode.text
        else:
            widgetname = None

        floatingNode = root.find("floating")
        if (floatingNode is not None) and (floatingNode.text is not None):
            floating = floatingNode.text == str(True)
        else:
            floating = True

        sharedDataWrite = {}
        sharedDataWriteNode = root.find("sharedDataWrite")
        if (sharedDataWriteNode is not None) and (sharedDataWriteNode.text is not None):
            for child in sharedDataWriteNode:
                if (child.get("datauid") is not None) and (child.get("signalName") is not None):
                    sharedDataWrite[child.get("datauid")] = child.get(
                        "signalName")

        if not len(sharedDataWrite):
            sharedDataWrite = None

        sharedDataRead = {}
        sharedDataReadNode = root.find("sharedDataRead")
        if (sharedDataReadNode is not None) and (sharedDataReadNode.text is not None):
            for child in sharedDataReadNode:
                if (child.get("datauid") is not None) and (child.get("slotName") is not None):
                    sharedDataRead[
                        child.get("datauid")] = child.get("slotName")

        if not len(sharedDataRead):
            sharedDataRead = None

        modelNode = root.find("model")
        if (modelNode is not None) and (modelNode.text is not None):
            model = modelNode.text
        else:
            model = None

        return PanelDescription(name, classname=classname, modulename=modulename, widgetname=widgetname,
                                floating=floating, sharedDataWrite=sharedDataWrite, sharedDataRead=sharedDataRead,
                                model=model)

    #=========================================================================
    # Properties
    #=========================================================================
    name = property(fget=getName, fset=setName)
    classname = property(fget=getClassname, fset=setClassname)
    modulename = property(fget=getModulename, fset=setModulename)
    widgetname = property(fget=getWidgetname, fset=setWidgetname)
    floating = property(fget=isFloating, fset=setFloating)
    sharedDataWrite = property(
        fget=getSharedDataWrite, fset=setSharedDataWrite)
    sharedDataRead = property(fget=getSharedDataRead, fset=setSharedDataRead)
    model = property(fget=getModel, fset=setModel)


class PanelDescription(TaurusGuiComponentDescription):
    '''
    A description of a taurusgui panel.
    This class is not a panel, but a container of the information required to
    build a panel.
    '''

    def __init__(self, *args, **kwargs):
        self.instrumentkey = kwargs.pop('instrumentkey', None)
        TaurusGuiComponentDescription.__init__(self, *args, **kwargs)

    @staticmethod
    def fromPanel(panel):
        name = str(panel.objectName())
        classname = panel.getWidgetClassName()
        modulename = panel.getWidgetModuleName()
        # in the case of classes known to the TaurusWidgetFactory,
        # do not store the modulename
        if modulename.startswith('taurus.') and \
           classname in TaurusWidgetFactory().getWidgetClassNames():
            modulename = None
        widgetname = None
        floating = panel.isFloating()
        sharedDataWrite = None
        sharedDataRead = None
        model = getattr(panel.widget(), 'model', None)
        if model is None or isinstance(model, basestring):
            pass
        elif hasattr(model, '__iter__'):
            # if model is a sequence, convert to space-separated string
            try:
                model = " ".join(model)
            except Exception, e:
                msg = ('Cannot convert %s to a space-separated string: %s' %
                       (model, e))
                Logger().debug(msg)
                model = None
        else:
            # ignore other "model" attributes (they are not from Taurus)
            model = None
        return PanelDescription(name, classname=classname,
                                modulename=modulename, widgetname=widgetname,
                                floating=floating,
                                sharedDataWrite=sharedDataWrite,
                                sharedDataRead=sharedDataRead, model=model)


class ToolBarDescription(TaurusGuiComponentDescription):
    '''
    A description of a toolbar to be inserted in a TaurusGUI.
    '''
    pass


class AppletDescription(TaurusGuiComponentDescription):
    '''
    A description of a widget to be inserted in the "applets bar" of the TaurusGUI.
    '''
    pass



# -*- coding: utf-8 -*-

#############################################################################
##
# This file is part of Taurus
##
# http://taurus-scada.org
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Taurus is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Taurus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
#############################################################################

"""Unit tests for UILoadable decorator"""

import os.path

from taurus.external import unittest
from taurus.external.qt import Qt
from taurus.qt.qtgui.util.ui import UILoadable
from taurus.qt.qtgui.test import BaseWidgetTestCase
from mywidget3 import MyWidget3


class UILoadableTestCase(unittest.TestCase):
    """
    Test cases for UILoadable decorator
    """

    @UILoadable
    class MyWidget1(Qt.QWidget):

        def __init__(self, parent=None):
            Qt.QWidget.__init__(self, parent)
            self.loadUi()
            self.my_button.setText("This is MY1 button")

    @UILoadable(with_ui="ui")
    class MyWidget2(Qt.QWidget):

        def __init__(self, parent=None):
            Qt.QWidget.__init__(self, parent)
            path = os.path.join(os.path.dirname(__file__), "ui", "mywidget2")
            self.loadUi(filename="mywidget2_custom.ui", path=path)
            self.ui.my_button.setText("This is MY2 button")

    def setUp(self):
        app = Qt.QApplication.instance()
        if app is None:
            app = Qt.QApplication([])
        self.__app = app

    def test_uiloadable_default(self):
        """Test UILoadable with default arguments"""
        widget = self.MyWidget1()
        self.assertEquals(widget.my_button.text(), "This is MY1 button",
                          "button text differs from expected")

    def test_uiloadable_customized(self):
        """Test UILoadable with customized filename and path"""
        widget = self.MyWidget2()
        self.assertTrue(hasattr(widget, "ui"),
                        "widget doesn't have 'ui' member")
        self.assertTrue(hasattr(widget.ui, "my_button"),
                        "widget.ui doesn't have a 'my_button' member")
        self.assertFalse(hasattr(widget, "my_button"),
                         "widget has a my_button member")
        self.assertEquals(widget.ui.my_button.text(), "This is MY2 button",
                          "button text differs from expected")


class Bug151_TestCase(BaseWidgetTestCase, unittest.TestCase):
    '''Test for bug 151: https://sourceforge.net/p/tauruslib/tickets/151/'''

    def test_bug151(self):
        '''Check inheritance of UILoadable classes across packages (bug #151)
        '''
        class Bug151_Widget(MyWidget3):
            pass
        try:
            Bug151_Widget()
        except:
            self.fail('Inheriting from UILoadable from another package fails')


def main():
    unittest.main()


if __name__ == "__main__":
    main()



#!/usr/bin/env python

#############################################################################
##
# This file is part of Taurus
##
# http://taurus-scada.org
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Taurus is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Taurus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
#############################################################################

"""This module contains a set of useful traceback elements based on python's
:mod:`traceback` system."""

import sys
import inspect
import traceback
import threading


def _get_thread(ident=None):
    if ident is None:
        return threading.current_thread()
    for th in threading.enumerate():
        if th.ident == ident:
            return th


def _get_frames():
    return sys._current_frames()


def format_frame_stacks(frames=None, limit=None):
    if frames is None:
        frames = _get_frames()
    frame_stacks = extract_frame_stacks(frames=frames, limit=limit)
    ret = []

    for ident, (frame, frame_stack) in frame_stacks.items():
        curr_th, th = _get_thread(), _get_thread(ident)
        if th is None:
            th_name = "<Unknown>"
            curr = ""
        else:
            th_name = th.name
            if curr_th.ident == th.ident:
                th_str = "(Current) "
            else:
                curr = ""
        ret.append("  Thread " + curr + th_name + " (" + str(ident) + ") in\n")
        format_stack = traceback.format_list(frame_stack)
        for i, line in enumerate(format_stack):
            line = "  " + line.replace("\n  ", "\n    ")
            ret.append(line)
    return ret


def extract_frame_stacks(frames=None, limit=None):
    if frames is None:
        frames = _get_frames()
    ret = {}
    for ident, frame in frames.items():
        frame_stack = traceback.extract_stack(frame, limit=limit)
        ret[ident] = frame, frame_stack
    return ret



#!/usr/bin/env python

#############################################################################
##
# This file is part of Taurus
##
# http://taurus-scada.org
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Taurus is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Taurus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
#############################################################################

"""This module provides a set of dialog based widgets"""

__all__ = ["TaurusInputDialog", "get_input"]

__docformat__ = 'restructuredtext'

from taurus.external.qt import Qt


def get_input(input_data, parent=None, input_panel_klass=None):
    """Static convenience function to get an input from the user using a
    dialog. The dialog will be modal.

    The input_data is a dictionary which contains information on how to build
    the input dialog. It **must** contains the following keys:

        - *prompt* <str>: message to be displayed

    The following are optional keys (and their corresponding default values):

        - *title* <str> (doesn't have default value)
        - *key* <str> (doesn't have default value):
          a label to be presented left to the input box represeting the label
        - *unit* <str> (doesn't have default value):
          a label to be presented right to the input box representing the units
        - *data_type* <str or sequence> ('String'):
          type of data to be requested. Standard
          accepted data types are 'String', 'Integer', 'Float', 'Boolean',
          'Text'. A list of elements will be interpreted as a selection.
          Default TaurusInputPanel class will interpret any custom data types as
          'String' and will display input widget accordingly. Custom
          data types can be handled differently by supplying a different
          input_panel_klass.
        - *minimum* <int/float> (-sys.maxint):
          minimum value (makes sence when data_type is 'Integer' or 'Float')
        - *maximum* <int/float> (sys.maxint):
          maximum value (makes sence when data_type is 'Integer' or 'Float')
        - *step* <int/float> (1):
          step size value (makes sence when data_type is 'Integer' or 'Float')
        - *decimals* <int> (1):
          number of decimal places to show (makes sence when data_type is
          'Float')
        - *default_value* <obj> (doesn't have default value):
          default value
        - *allow_multiple* <bool> (False):
          allow more than one value to be selected (makes sence when data_type
          is a sequence of possibilities)

    :param input_data: a dictionary with information on how to build the input dialog
    :type input_data: :py:obj:`dict`
    :param parent: parent widget
    :type parent: PyQt4.QtGui.QWidget
    :param input_panel_klass: python class to be used as input panel [default: :class:`~taurus.qt.qtgui.panel.TaurusInputPanel`]
    :type input_panel_klass: :class:`~taurus.qt.qtgui.panel.TaurusInputPanel`

    :return: a tuple containing value selected and boolean which is true if
             user accepted the dialog (pressed Ok) or false otherwise
    :rtype: tuple< obj, bool >

    Examples::

        d1 = dict(prompt="What's your name?", data_type="String")
        d2 = dict(prompt="What's your age?", data_type="Integer",
                  default_value=4, maximum=100, key="Age", unit="years")
        d3 = dict(prompt="What's your favourite number?", data_type="Float",
                  default_value=0.1, maximum=88.8, key="Number")
        d4 = dict(prompt="What's your favourite car brand?",
                  data_type=["Mazda", "Skoda", "Citroen", "Mercedes", "Audi", "Ferrari"],
                  default_value="Mercedes")
        d5 = dict(prompt="Select some car brands", allow_multiple=True,
                  data_type=["Mazda", "Skoda", "Citroen", "Mercedes", "Audi", "Ferrari"],
                  default_value=["Mercedes", "Citroen"])
        d6 = dict(prompt="What's your favourite color?", key="Color",
                  data_type=["blue", "red", "green"], default_value="red")
        d7 = dict(prompt="Do you like bears?",
                  data_type='Boolean', key="Yes/No", default_value=True)
        d8 = dict(prompt="Please write your memo",
                  data_type='Text', key="Memo", default_value="By default a memo is a long thing")
        for d in [d1, d2, d3, d4, d5, d6, d7, d8]:
            get_input(input_data=d, title=d['prompt'])

    """

    if input_panel_klass is None:
        from taurus.qt.qtgui.panel import TaurusInputPanel
        input_panel_klass = TaurusInputPanel
    dialog = TaurusInputDialog(input_data=input_data, parent=parent,
                               input_panel_klass=input_panel_klass)
    dialog.exec_()
    return dialog.value(), dialog.result()


class TaurusInputDialog(Qt.QDialog):
    """The TaurusInputDialog class provides a simple convenience dialog to get
    a single value from the user.
    """

    def __init__(self, input_data=None, parent=None,
                 input_panel_klass=None, designMode=False):
        if input_panel_klass is None:
            from taurus.qt.qtgui.panel import TaurusInputPanel
            input_panel_klass = TaurusInputPanel
        self.input_data = input_data
        self.input_panel_klass = input_panel_klass
        Qt.QDialog.__init__(self, parent)
        if input_data and 'title' in input_data:
            self.setWindowTitle(input_data['title'])
        layout = Qt.QVBoxLayout()
        self.setLayout(layout)
        self._panel = panel = input_panel_klass(input_data, self)
        layout.setContentsMargins(0, 0, 0, 0)
        layout.addWidget(self._panel)
        panel.buttonBox().accepted.connect(self.accept)
        panel.buttonBox().rejected.connect(self.reject)
        self._panel.setInputFocus()

    def panel(self):
        """Returns the :class:`taurus.qt.qtgui.panel.TaurusInputPanel`.

        :return: the internal panel
        :rtype: taurus.qt.qtgui.panel.TaurusInputPanel"""
        return self._panel

    def value(self):
        """Returns the value selected by the user.

        :return: the value selected by the user"""
        return self.panel().value()


def demo():
    """Input Dialog"""

    input_data = dict(prompt="What's your favourite car brand?",
                      data_type=["Mazda", "Skoda", "Citroen",
                                 "Mercedes", "Audi", "Ferrari"],
                      default_value="Mercedes")
    w = TaurusInputDialog(input_data=input_data)
    w.show()
    return w


def main():
    import taurus.qt.qtgui.application

    Application = taurus.qt.qtgui.application.TaurusApplication

    app = Application.instance()
    owns_app = app is None

    if owns_app:
        app = Qt.QApplication([])
        app.setApplicationName("Taurus input dialog demo")
        app.setApplicationVersion("1.0")

    d1 = dict(prompt="What's your name?", data_type="String")
    d2 = dict(prompt="What's your age?", data_type="Integer",
              default_value=4, maximum=100, key="Age", unit="years")
    d3 = dict(prompt="What's your favourite number?", data_type="Float",
              default_value=0.1, maximum=88.8, key="Number")
    d4 = dict(prompt="What's your favourite car brand?",
              data_type=["Mazda", "Skoda", "Citroen",
                         "Mercedes", "Audi", "Ferrari"],
              default_value="Mercedes")
    d5 = dict(prompt="Select some car brands", allow_multiple=True,
              data_type=["Mazda", "Skoda", "Citroen",
                         "Mercedes", "Audi", "Ferrari"],
              default_value=["Mercedes", "Citroen"])
    d6 = dict(prompt="What's your favourite color?", key="Color",
              data_type=["blue", "red", "green"], default_value="red")
    d7 = dict(prompt="Do you like bears?",
              data_type='Boolean', key="Yes/No", default_value=True)
    d8 = dict(prompt="Please write your memo",
              data_type='Text', key="Memo", default_value="By default a memo is\na long thing")

    for d in [d1, d2, d3, d4, d5, d6, d7, d8]:
        print get_input(input_data=d, title=d['prompt'])

if __name__ == "__main__":
    main()



#!/usr/bin/env python

#############################################################################
##
# This file is part of Taurus
##
# http://taurus-scada.org
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Taurus is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Taurus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
#############################################################################

"""Test for taurus.core.tango.test.test_tangovalidator..."""

# __all__ = []

import numpy
from taurus.external import unittest
from taurus.external.pint import Quantity
import taurus
from taurus.test import insertTest
from taurus.core.taurusbasetypes import DataType
from taurus.core.evaluation.evalattribute import EvaluationAttrValue


@insertTest(helper_name='read_attr',
            attr_fullname='eval:1',
            expected=dict(rvalue=Quantity(1, 'dimensionless'),
                          type=DataType.Integer,
                          label='1',
                          writable=False,
                          ),
            expected_attrv=dict(rvalue=Quantity([], 'dimensionless'),
                                wvalue=None
                                ),
            expectedshape=None
            )
@insertTest(helper_name='read_attr',
            attr_fullname='eval:Quantity("1m")+Quantity("2m")',
            expected=dict(rvalue=Quantity(3, 'm'),
                          type=DataType.Integer,
                          label='Quantity("1m")+Quantity("2m")',
                          writable=False,
                          ),
            expected_attrv=dict(rvalue=Quantity(3, 'm'),
                                wvalue=None
                                ),
            expectedshape=None
            )
@insertTest(helper_name='read_attr',
            attr_fullname='eval:foo=Quantity("1m");bar=Quantity("2m");foo+bar',
            expected=dict(rvalue=Quantity(3, 'm'),
                          type=DataType.Integer,
                          label='foo+bar',
                          writable=False,
                          ),
            expected_attrv=dict(rvalue=Quantity(3, 'm'),
                                wvalue=None
                                ),
            expectedshape=None
            )
@insertTest(helper_name='read_attr',
            attr_fullname='eval:foo=Q("1m");foo-Q("2m")',
            expected=dict(rvalue=Quantity(-1, 'm'),
                          type=DataType.Integer,
                          label='foo-Q("2m")',
                          writable=False,
                          ),
            expected_attrv=dict(rvalue=Quantity(-1, 'm'),
                                wvalue=None
                                ),
            expectedshape=None
            )
@insertTest(helper_name='read_attr',
            attr_fullname='eval:[123, 123, 123]',
            expected=dict(rvalue=Quantity([123, 123, 123], 'dimensionless'),
                          type=DataType.Integer,
                          label='[123, 123, 123]',
                          writable=False,
                          ),
            expected_attrv=dict(rvalue=Quantity([123, 123, 123],
                                                'dimensionless'),
                                wvalue=None
                                ),
            expectedshape=(3, )
            )
@insertTest(helper_name='read_attr',
            attr_fullname='eval:Quantity([1.23, 1.23, 1.23], "m")',
            expected=dict(rvalue=Quantity([1.23, 1.23, 1.23], 'm'),
                          type=DataType.Float,
                          label='Quantity([1.23, 1.23, 1.23], "m")',
                          writable=False,
                          ),
            expected_attrv=dict(rvalue=Quantity([1.23, 1.23, 1.23], 'm'),
                                wvalue=None
                                ),
            expectedshape=(3, )
            )
@insertTest(helper_name='read_attr',
            attr_fullname='eval:foo=Quantity([1.23, 1.23, 1.23], "m");foo*10',
            expected=dict(rvalue=Quantity([12.3, 12.3, 12.3], 'm'),
                          type=DataType.Float,
                          label='foo*10',
                          writable=False,
                          ),
            expected_attrv=dict(rvalue=Quantity([12.3, 12.3, 12.3],
                                                'm'),
                                wvalue=None
                                ),
            expectedshape=(3, )
            )
@insertTest(helper_name='read_attr',
            attr_fullname='eval:[[1,2,3]]*3',
            expected=dict(rvalue=Quantity([[1, 2, 3]] * 3, 'dimensionless'),
                          type=DataType.Integer,
                          label='[[1,2,3]]*3',
                          writable=False,
                          ),
            expected_attrv=dict(rvalue=Quantity([[1, 2, 3]] * 3,
                                                'dimensionless'),
                                wvalue=None
                                ),
            expectedshape=(3, 3)
            )
@insertTest(helper_name='read_attr',
            attr_fullname='eval:foo=1;foo+{eval:foo=2;foo}+{eval:foo=3;foo}',
            expected=dict(rvalue=Quantity(6, 'dimensionless'),
                          type=DataType.Integer,
                          label='foo+foo+foo',
                          writable=False,
                          ),
            expected_attrv=dict(rvalue=Quantity(6, 'dimensionless'),
                                wvalue=None
                                ),
            expectedshape=None
            )
@insertTest(helper_name='read_attr',
            attr_fullname='eval:True',
            expected=dict(rvalue=True,
                          type=DataType.Boolean,
                          label='True',
                          writable=False,
                          ),
            expected_attrv=dict(rvalue=True,
                                wvalue=None
                                ),
            expectedshape=None
            )
@insertTest(helper_name='read_attr',
            attr_fullname='eval:foo=1;bar=2;foo==bar',
            expected=dict(rvalue=False,
                          type=DataType.Boolean,
                          label='foo==bar',
                          writable=False,
                          ),
            expected_attrv=dict(rvalue=False,
                                wvalue=None
                                ),
            expectedshape=None
            )
@insertTest(helper_name='read_attr',
            attr_fullname='eval:[[True, False, True]]*3',
            expected=dict(rvalue=[[True, False, True]] * 3,
                          type=DataType.Boolean,
                          label='[[True, False, True]]*3',
                          writable=False,
                          ),
            expected_attrv=dict(rvalue=[[True, False, True]] * 3,
                                wvalue=None
                                ),
            expectedshape=(3, 3)
            )
@insertTest(helper_name='read_attr',
            attr_fullname='eval:"1m"',
            expected=dict(rvalue="1m",
                          type=DataType.String,
                          label='"1m"',
                          writable=False,
                          ),
            expected_attrv=dict(rvalue="1m",
                                wvalue=None
                                ),
            expectedshape=None
            )
@insertTest(helper_name='read_attr',
            attr_fullname='eval:"m"*3',
            expected=dict(rvalue="mmm",
                          type=DataType.String,
                          label='"m"*3',
                          writable=False,
                          ),
            expected_attrv=dict(rvalue="mmm",
                                wvalue=None
                                ),
            expectedshape=None
            )
@insertTest(helper_name='read_attr',
            attr_fullname='eval:"2+5-10"',
            expected=dict(rvalue="2+5-10",
                          type=DataType.String,
                          label='"2+5-10"',
                          writable=False,
                          ),
            expected_attrv=dict(rvalue="2+5-10",
                                wvalue=None
                                ),
            expectedshape=None
            )
@insertTest(helper_name='read_attr',
            attr_fullname='eval:[["HelloWorld", "HelloWorld", "HelloWorld"]]*3',
            expected=dict(rvalue=[["HelloWorld", "HelloWorld", "HelloWorld"]] * 3,
                          type=DataType.String,
                          label='[["HelloWorld", "HelloWorld",' +
                                ' "HelloWorld"]]*3',
                          writable=False,
                          ),
            expected_attrv=dict(rvalue=[["HelloWorld", "HelloWorld",
                                         "HelloWorld"]] * 3,
                                wvalue=None
                                ),
            expectedshape=(3, 3)
            )
class EvalAttributeTestCase(unittest.TestCase):

    def read_attr(self, attr_fullname, expected={}, expected_attrv={},
                  expectedshape=None):
        """check creation and correct read of an evaluationAttribute"""

        a = taurus.Attribute(attr_fullname)
        read_value = a.read()

        msg = ('read() for "{0}" did not return an EvaluationAttrValue ' +
               '(got a {1})').format(attr_fullname,
                                     read_value.__class__.__name__)
        self.assertTrue(isinstance(read_value, EvaluationAttrValue), msg)

        # Test attribute
        for k, exp in expected.iteritems():
            try:
                got = getattr(a, k)
            except AttributeError:
                msg = ('The attribute, "%s" does not provide info on %s' %
                       (attr_fullname, k))
                self.fail(msg)
            msg = ('%s for "%s" should be %r (got %r)' %
                   (attr_fullname, k,  exp, got))
            self.__assertValidValue(exp, got, msg)

        # Test attribute value
        for k, exp in expected_attrv.iteritems():
            try:
                got = getattr(read_value, k)
            except AttributeError:
                msg = ('The read value for "%s" does not provide info on %s' %
                       (attr_fullname, k))
                self.fail(msg)
            msg = ('%s for "%s" should be %r (got %r)' %
                   (attr_fullname, k,  exp, got))
            self.__assertValidValue(exp, got, msg)

        if expectedshape is not None:
            shape = numpy.shape(read_value.rvalue)
            msg = ('rvalue.shape for %s should be %r (got %r)' %
                   (attr_fullname, expectedshape, shape))
            self.assertEqual(shape, expectedshape, msg)

    def __assertValidValue(self, exp, got, msg):
        # if we are dealing with quantities, use the magnitude for comparing
        if isinstance(got, Quantity):
            got = got.to(Quantity(exp).units).magnitude
        if isinstance(exp, Quantity):
            exp = exp.magnitude
        try:
            # for those values that can be handled by numpy.allclose()
            chk = numpy.allclose(got, exp)
        except:
            # for the rest
            if isinstance(got, numpy.ndarray):
                got = got.tolist()
            chk = bool(got == exp)

        self.assertTrue(chk, msg)



#!/usr/bin/env python

#############################################################################
##
# This file is part of Taurus
##
# http://taurus-scada.org
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Taurus is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Taurus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
#############################################################################

'''Utility functions to deal with non-ideal (fuzzy) tests'''


def loopTest(testname, maxtries=100, maxfails=10):
    '''Run a test `maxtries` times or until it fails `maxfails` times and
    report the number of tries and failures.

    :param testname: (str) test name. see:
                     :meth:`unittest.TestLoader.loadTestsFromName`
    :param maxtries: (int) maximum number of runs
    :param maxfails: (int) maximum number of failed runs

    :return: (tuple) a tuple of ints: tries, failures
    '''
    from taurus.external import unittest
    suite = unittest.defaultTestLoader.loadTestsFromName(testname)
    runner = unittest.TextTestRunner(verbosity=0)
    i, f = 0, 0
    while f < maxfails and i < maxtries:
        i += 1
        result = runner.run(suite)
        if not result.wasSuccessful():
            f += 1
    return i, f


def loopSubprocess(target, maxtries=100, maxfails=10, okvalues=(0,), args=(),
                   kwargs=None):
    '''Run a callable as a subprocess `maxtries` times or until it fails
    `maxfails` times and report the number of tries and failures.
    The callable is run as a subprocess and it is considered to run fine if
    the subprocess exit code is in the okValues list.

    :param target: (callable) a callable test
    :param maxtries: (int) maximum number of runs
    :param maxfails: (int) maximum number of failed runs
    :param okvalues: (seq) a sequence containing exit values of cmd which
                     are considered to be successful runs.
    :param args: (seq) arguments for running the target function
    :param kwargs: (dict) keyword arguments for running the target function


    :return: (tuple) a tuple of ints: tries, failures
    '''
    if kwargs is None:
        kwargs = {}
    from multiprocessing import Process
    i, f = 0, 0
    while f < maxfails and i < maxtries:
        i += 1
        p = Process(target=target, args=args, kwargs=kwargs)
        p.start()
        p.join()
        if p.exitcode not in okvalues:
            f += 1
    return i, f


def calculateTestFuzziness(test, maxtries=100, maxfails=10, **kwargs):
    '''Estimate the fuzziness of a test by running it many times and counting
    the failures. In this context, we assume that there is an underlying
    problem and but that the test is not perfect and only fails (triggers the
    problem) with a certain failure rate.

    :param testname: (str) test name. see:
                     :meth:`unittest.TestLoader.loadTestsFromName`
    :param maxtries: (int) maximum number of runs
    :param maxfails: (int) maximum number of failed runs

    :return: (tuple) a tuple (f,df,n) where f is the failure rate, df is its
             standard deviation, and n is the number of consecutive
             times that the test should be passed to have a confidence>99%%
             that the bug is fixed'
    '''
    print ("Running the test %i times (or until it fails %i times)" +
           "to estimate the failure rate") % (maxtries, maxfails)
    import numpy

    if isinstance(test, str):
        tries, fails = loopTest(test, maxtries=maxtries, maxfails=maxfails)
    else:
        tries, fails = loopSubprocess(test, maxtries=maxtries,
                                      maxfails=maxfails, **kwargs)
    r = float(fails) / tries
    dr = numpy.sqrt(fails) / tries
    print 'Failure rate = %g +/- %g  (%i/%i)' % (r, dr, fails, tries)
    # calculating n using p-value=1% and failure rate with -1 sigma
    n = numpy.ceil(numpy.log(.01) / numpy.log(1 - (r - dr)))
    print ('Number of consecutive times that the test should be passed ' +
           'to have a confidence>99%% that the bug is fixed: %g') % n
    return r, dr, n


if __name__ == "__main__":

    def kk():
        # exits with fail 1/3 of the times
        from numpy.random import randint, seed
        seed()
        k = randint(3)
        if not k:
            exit(1)
        return

    print calculateTestFuzziness(kk)

#     print calculateTestFuzziness('test_pytango_bug659.TestPyTango_Bug659')
#
#   _, f = loopTest('test_pytango_bug659.TestPyTango_Bug659',
#                  maxtries=100, maxfails=1)



#!/usr/bin/env python

#############################################################################
##
# This file is part of Taurus
##
# http://taurus-scada.org
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Taurus is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Taurus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
#############################################################################

"""
TaurusDevicePanel.py:
"""

__all__ = ["TaurusDevicePanel", "TaurusDevPanel"]

__docformat__ = 'restructuredtext'

import re
import traceback

import taurus
from taurus.external.qt import Qt

from taurus import tauruscustomsettings
from taurus.core.taurusbasetypes import TaurusDevState, TaurusElementType
from taurus.core.taurusattribute import TaurusAttribute
from taurus.core.taurusdevice import TaurusDevice
from taurus.qt.qtgui.container import TaurusWidget, TaurusMainWindow
from taurus.qt.qtgui.display import TaurusLabel
from taurus.qt.qtgui.display import TaurusLed
from taurus.qt.qtgui.panel.taurusform import TaurusForm
from taurus.qt.qtgui.panel.taurusform import TaurusCommandsForm
from taurus.qt.qtgui.util.ui import UILoadable
from taurus.qt.qtgui.icon import getCachedPixmap

from taurus.core.tango.tangodatabase import TangoDevInfo  # @todo: Tango-centric!

###############################################################################
# TaurusDevicePanel (from Vacca)

# Variables that control TaurusDevicePanel shape

STATUS_HEIGHT = 170
SPLIT_SIZES = [15, 65, 20]
IMAGE_SIZE = (200, 100)  # (width,height)

# Helper methods


def matchCl(m, k):  # TODO: Tango-centric
    return re.match(m.lower(), k.lower())


def searchCl(m, k):  # TODO: Tango-centric
    if m.startswith('^') or m.startswith('(^') or '(?!^' in m:
        return matchCl(m, k)
    return re.search(m.lower(), k.lower())


def get_regexp_dict(dct, key, default=None):  # TODO: Tango-centric
    for k, v in dct.items():  # Trying regular expression match
        if matchCl(k, key):
            return v
    for k, v in dct.items():  # If failed, trying if key is contained
        if k.lower() in key.lower():
            return v
    if default is not None:
        return default
    else:
        raise Exception('KeyNotFound:%s' % k)


def get_eqtype(dev):  # TODO: Tango-centric
    ''' It extracts the eqtype from a device name like domain/family/eqtype-serial'''
    try:
        eq = str(dev).split('/')[-1].split('-', 1)[0].upper()
    except:
        eq = ''
    return eq


def str_to_filter(seq):  # TODO: Tango-centric
    try:
        f = eval(seq)
    except:
        f = seq
    if isinstance(f, basestring):
        return {'.*': [f]}
    elif isinstance(f, list):
        return {'.*': f}
    else:
        return f

# Stacked palette


def get_White_palette():
    palette = Qt.QPalette()

    brush = Qt.QBrush(Qt.QColor(255, 255, 255))
    brush.setStyle(Qt.Qt.SolidPattern)
    palette.setBrush(Qt.QPalette.Active, Qt.QPalette.Base, brush)

    brush = Qt.QBrush(Qt.QColor(255, 255, 255))
    brush.setStyle(Qt.Qt.SolidPattern)
    palette.setBrush(Qt.QPalette.Active, Qt.QPalette.Window, brush)

    brush = Qt.QBrush(Qt.QColor(255, 255, 255))
    brush.setStyle(Qt.Qt.SolidPattern)
    palette.setBrush(Qt.QPalette.Inactive, Qt.QPalette.Base, brush)

    brush = Qt.QBrush(Qt.QColor(255, 255, 255))
    brush.setStyle(Qt.Qt.SolidPattern)
    palette.setBrush(Qt.QPalette.Inactive, Qt.QPalette.Window, brush)

    brush = Qt.QBrush(Qt.QColor(255, 255, 255))
    brush.setStyle(Qt.Qt.SolidPattern)
    palette.setBrush(Qt.QPalette.Disabled, Qt.QPalette.Base, brush)

    brush = Qt.QBrush(Qt.QColor(255, 255, 255))
    brush.setStyle(Qt.Qt.SolidPattern)
    palette.setBrush(Qt.QPalette.Disabled, Qt.QPalette.Window, brush)
    return palette

# TaurusDevicePanel class


class TaurusDevicePanel(TaurusWidget):
    '''
    TaurusDevPanel is a Taurus Application inspired in Jive and Atk Panel.

    It Provides a Device selector and a panel for displaying information from
    the selected device.
    '''
    # TODO: Tango-centric (This whole class should be called TangoDevicePanel)
    # TODO: a scheme-agnostic base class should be implemented

    READ_ONLY = False
    # A dictionary like {device_regexp:[attribute_regexps]}
    _attribute_filter = {}
    # A dictionary like {device_regexp:[(command_regexp,default_args)]}
    _command_filter = {}
    _icon_map = {}  # A dictionary like {device_regexp:pixmap_url}

    @classmethod
    def setIconMap(klass, filters):
        """A dictionary like {device_regexp:pixmap_url}"""
        klass._icon_map = filters

    @classmethod
    def getIconMap(klass):
        return klass._icon_map

    @classmethod
    def setAttributeFilters(klass, filters):
        """
        It will set the attribute filters
        filters will be like: {device_regexp:[attribute_regexps]}
        example: {'.*/VGCT-.*': ['ChannelState','p[0-9]']}
        """
        klass._attribute_filter.update(filters)

    @classmethod
    def getAttributeFilters(klass):
        return klass._attribute_filter

    @classmethod
    def setCommandFilters(klass, filters):
        """
        It will set the command filters
        filters will be like: {device_regexp:[command_regexps]}
        example::

          {'.*/IPCT-.*': (
                           ('setmode',('SERIAL','LOCAL','STEP','FIXED','START','PROTECT')),
                           ('onhv1',()), ('offhv1',()), ('onhv2',()), ('offhv2',()),
                           ('sendcommand',())
                         ),}

        """
        klass._command_filter.update(filters)

    @classmethod
    def getCommandFilters(klass):
        return klass._command_filter

    ###########################################################################

    def __init__(self, parent=None, model=None, palette=None, bound=True):
        TaurusWidget.__init__(self, parent)
        if palette:
            self.setPalette(palette)
        self.setLayout(Qt.QGridLayout())
        self.bound = bound
        self._dups = []

        self.setWindowTitle('TaurusDevicePanel')
        self._label = Qt.QLabel()
        self._label.font().setBold(True)

        self._stateframe = TaurusWidget(self)
        self._stateframe.setLayout(Qt.QGridLayout())
        self._stateframe.layout().addWidget(Qt.QLabel('State'), 0, 0, Qt.Qt.AlignCenter)
        self._statelabel = TaurusLabel(self._stateframe)
        self._statelabel.setMinimumWidth(100)
        self._statelabel.setBgRole('state')
        self._stateframe.layout().addWidget(self._statelabel, 0, 1, Qt.Qt.AlignCenter)
        self._state = TaurusLed(self._stateframe)
        self._state.setShowQuality(False)
        self._stateframe.layout().addWidget(self._state, 0, 2, Qt.Qt.AlignCenter)

        self._statusframe = Qt.QScrollArea(self)
        self._status = TaurusLabel(self._statusframe)
        self._status.setBgRole('none')
        self._status.setAlignment(Qt.Qt.AlignLeft)
        self._status.setFixedHeight(2000)
        self._status.setFixedWidth(5000)
        # self._statusframe.setFixedHeight(STATUS_HEIGHT)
        self._statusframe.setHorizontalScrollBarPolicy(Qt.Qt.ScrollBarAlwaysOn)
        self._statusframe.setVerticalScrollBarPolicy(Qt.Qt.ScrollBarAlwaysOn)
        self._statusframe.setWidget(self._status)
        self._statusframe.setPalette(get_White_palette())

        self._attrsframe = Qt.QTabWidget(self)

        # Horizontal will not allow to show labels of attributes!
        self._splitter = Qt.QSplitter(Qt.Qt.Vertical, self)

        self._attrs, self._comms = None, None

        self.layout().addWidget(self._splitter, 0, 0)
        self._header = Qt.QFrame()
        self._header.setFixedHeight(1.1 * IMAGE_SIZE[1])
        self._header.setLayout(Qt.QGridLayout())

        self._dup = Qt.QPushButton()
        qpixmap = Qt.QIcon("actions:window-new.svg")
        self._dup.setIcon(Qt.QIcon(qpixmap))
        self._dup.setIconSize(Qt.QSize(15, 15))
        self._dup.pressed.connect(self.duplicate)

        self._image = Qt.QLabel()

        self._header.layout().addWidget(self._image, 0, 0, 2, 1, Qt.Qt.AlignCenter)
        self._header.layout().addWidget(self._label, 0, 1, Qt.Qt.AlignLeft)
        self._header.layout().addWidget(self._stateframe, 1, 1, 1, 2, Qt.Qt.AlignLeft)
        self._header.layout().addWidget(self._dup, 0, 2, Qt.Qt.AlignRight)

        self._splitter.insertWidget(0, self._header)
        self._splitter.insertWidget(1, self._attrsframe)
        self._splitter.insertWidget(2, self._statusframe)
        self._splitter.setSizes(SPLIT_SIZES)
        [self._splitter.setStretchFactor(i, v)
         for i, v in enumerate(SPLIT_SIZES)]
        self._splitter.setCollapsible(0, False)
        self._splitter.setCollapsible(1, False)

        if model:
            self.setModel(model)

    def loadConfigFile(self, ifile=None):
        self.info('In TaurusDevicePanel.loadConfigFile(%s)' % ifile)
        if isinstance(ifile, file) or isinstance(ifile, str) and not ifile.endswith('.py'):
            TaurusWidget.loadConfigFile(self, ifile)
        else:
            from imp import load_source
            config_file = load_source('config_file', ifile)
            af, cf, im = [getattr(config_file, x, None) for x in (
                'AttributeFilters', 'CommandFilters', 'IconMap')]
            if af is not None:
                self.setAttributeFilters(af)
            if cf is not None:
                self.setCommandFilters(cf)
            if im is not None:
                self.setIconMap(im)
        self.debug('AttributeFilters are:\n%s' % self.getAttributeFilters())

    def duplicate(self):
        self._dups.append(TaurusDevicePanel(bound=False))
        self._dups[-1].setModel(self.getModel())
        self._dups[-1].show()

    @Qt.pyqtSlot('QString')
    def setModel(self, model, pixmap=None):
        model, modelclass, raw = str(model).strip(), '', model
        if model:
            model = model and model.split()[0] or ''
            modelclass = taurus.Factory().findObjectClass(model)
        self.trace('In TaurusDevicePanel.setModel(%s(%s),%s)' %
                   (raw, modelclass, pixmap))
        if model == self.getModel():
            return
        elif raw is None or not model or not modelclass:
            if self.getModel():
                self.detach()
            return
        elif issubclass(modelclass, TaurusAttribute):
            # if model.lower().endswith('/state'):
            model = model.rsplit('/', 1)[0]  # TODO: Tango-centric
        elif not issubclass(modelclass, TaurusDevice):
            self.warning('TaurusDevicePanel accepts only Device models')
            return
        try:
            taurus.Device(model).ping()  # TODO: Tango-centric
            if self.getModel():
                self.detach()  # Do not dettach previous model before pinging the new one (fail message will be shown at except: clause)
            TaurusWidget.setModel(self, model)
            self.setWindowTitle(str(model).upper())
            model = self.getModel()
            self._label.setText(model.upper())
            font = self._label.font()
            font.setPointSize(15)
            self._label.setFont(font)
            if pixmap is None and self.getIconMap():
                for k, v in self.getIconMap().items():
                    if searchCl(k, model):
                        pixmap = v
            if pixmap is not None:
                # print 'Pixmap is %s'%pixmap
                qpixmap = Qt.QPixmap(pixmap)
                if qpixmap.height() > .9 * IMAGE_SIZE[1]:
                    qpixmap = qpixmap.scaledToHeight(.9 * IMAGE_SIZE[1])
                if qpixmap.width() > .9 * IMAGE_SIZE[0]:
                    qpixmap = qpixmap.scaledToWidth(.9 * IMAGE_SIZE[0])
            else:
                logo = getattr(tauruscustomsettings, 'ORGANIZATION_LOGO',
                               "logos:taurus.png")
                qpixmap = getCachedPixmap(logo)

            self._image.setPixmap(qpixmap)
            self._state.setModel(model + '/state')  # TODO: Tango-centric
            if hasattr(self, '_statelabel'):
                self._statelabel.setModel(
                    model + '/state')  # TODO: Tango-centric
            self._status.setModel(model + '/status')  # TODO: Tango-centric
            try:
                self._attrsframe.clear()
                filters = get_regexp_dict(
                    TaurusDevicePanel._attribute_filter, model, ['.*'])
                if hasattr(filters, 'keys'):
                    filters = filters.items()  # Dictionary!
                if filters and isinstance(filters[0], (list, tuple)):  # Mapping
                    self._attrs = []
                    for tab, attrs in filters:
                        self._attrs.append(self.get_attrs_form(
                            device=model, filters=attrs, parent=self))
                        self._attrsframe.addTab(self._attrs[-1], tab)
                else:
                    if self._attrs and isinstance(self._attrs, list):
                        self._attrs = self._attrs[0]
                    self._attrs = self.get_attrs_form(
                        device=model, form=self._attrs, filters=filters, parent=self)
                    if self._attrs:
                        self._attrsframe.addTab(self._attrs, 'Attributes')
                if not TaurusDevicePanel.READ_ONLY:
                    self._comms = self.get_comms_form(model, self._comms, self)
                    if self._comms:
                        self._attrsframe.addTab(self._comms, 'Commands')
                if SPLIT_SIZES:
                    self._splitter.setSizes(SPLIT_SIZES)
            except:
                self.warning(traceback.format_exc())
                qmsg = Qt.QMessageBox(Qt.QMessageBox.Critical, '%s Error' %
                                      model, '%s not available' % model, Qt.QMessageBox.Ok, self)
                qmsg.setDetailedText(traceback.format_exc())
                qmsg.show()
        except:
            self.warning(traceback.format_exc())
            qmsg = Qt.QMessageBox(Qt.QMessageBox.Critical, '%s Error' %
                                  model, '%s not available' % model, Qt.QMessageBox.Ok, self)
            qmsg.show()
        self.setWindowTitle(self.getModel())
        return

    def detach(self):
        self.trace('In TaurusDevicePanel(%s).detach()' % self.getModel())
        _detached = []
        # long imports to avoid comparison problems in the isinstance below
        import taurus.qt.qtgui.container
        import taurus.qt.qtgui.base

        def detach_recursive(obj):
            if obj in _detached:
                return
            if isinstance(obj, taurus.qt.qtgui.container.TaurusBaseContainer):
                for t in obj.taurusChildren():
                    detach_recursive(t)
            if obj is not self and isinstance(obj, taurus.qt.qtgui.base.TaurusBaseWidget):
                try:
                    if getattr(obj, 'model', None):
                        #self.debug('detaching %s from %s'%(obj,obj.model))
                        obj.setModel([] if isinstance(obj, TaurusForm) else '')
                except:
                    self.warning('detach of %s failed!' % obj)
                    self.warning(traceback.format_exc())
            _detached.append(obj)
        detach_recursive(self)
        try:
            self._label.setText('')
            self._state.setModel('')
            if hasattr(self, '_statelabel'):
                self._statelabel.setModel('')
            self._status.setModel('')
            self._image.setPixmap(Qt.QPixmap())
        except:
            self.warning(traceback.format_exc())

    def get_attrs_form(self, device, form=None, filters=None, parent=None):
        filters = filters or get_regexp_dict(
            TaurusDevicePanel._attribute_filter, device, ['.*'])
        self.trace('In TaurusDevicePanel.get_attrs_form(%s,%s)' %
                   (device, filters))
        allattrs = sorted(str(a) for a in taurus.Device(
            device).get_attribute_list() if str(a).lower() not in ('state', 'status'))
        attrs = []
        for a in filters:
            for t in allattrs:
                if a and searchCl(a.strip(), t.strip()):
                    aname = '%s/%s' % (device, t)
                    if aname not in attrs:
                        attrs.append(aname)
        if attrs:
            #self.trace( 'Matching attributes are: %s' % str(attrs)[:100])
            if form is None:
                form = TaurusForm(parent)
            elif hasattr(form, 'setModel'):
                form.setModel([])
            # Configuring the TauForm:
            form.setWithButtons(False)
            form.setWindowTitle(device)
            try:
                form.setModel(attrs)
            except Exception:
                self.warning(
                    'TaurusDevicePanel.ERROR: Unable to setModel for TaurusDevicePanel.attrs_form!!: %s' % traceback.format_exc())
            return form
        else:
            return None

    def get_comms_form(self, device, form=None, parent=None):
        self.trace('In TaurusDevicePanel.get_comms_form(%s)' % device)
        params = get_regexp_dict(TaurusDevicePanel._command_filter, device, [])
        # If filters are defined only listed devices will show commands
        if TaurusDevicePanel._command_filter and not params:
            self.debug(
                'TaurusDevicePanel.get_comms_form(%s): By default an unknown device type will display no commands' % device)
            return None
        if not form:
            form = TaurusCommandsForm(parent)
        elif hasattr(form, 'setModel'):
            form.setModel('')
        try:
            form.setModel(device)
            if params:
                form.setSortKey(lambda x, vals=[s[0].lower() for s in params]: vals.index(
                    x.cmd_name.lower()) if str(x.cmd_name).lower() in vals else 100)
                form.setViewFilters([lambda c: str(c.cmd_name).lower() not in (
                    'state', 'status') and any(searchCl(s[0], str(c.cmd_name)) for s in params)])
                form.setDefaultParameters(dict((k, v) for k, v in (
                    params if not hasattr(params, 'items') else params.items()) if v))
            for wid in form._cmdWidgets:
                if not hasattr(wid, 'getCommand') or not hasattr(wid, 'setDangerMessage'):
                    continue
                if re.match('.*(on|off|init|open|close).*', str(wid.getCommand().lower())):
                    wid.setDangerMessage(
                        'This action may affect other systems!')
            # form._splitter.setStretchFactor(1,70)
            # form._splitter.setStretchFactor(0,30)
            form._splitter.setSizes([80, 20])
        except Exception:
            self.warning(
                'Unable to setModel for TaurusDevicePanel.comms_form!!: %s' % traceback.format_exc())
        return form


def filterNonExported(obj):
    if not isinstance(obj, TangoDevInfo) or obj.exported():
        return obj
    return None


@UILoadable(with_ui='_ui')
class TaurusDevPanel(TaurusMainWindow):
    '''
    TaurusDevPanel is a Taurus Application inspired in Jive and Atk Panel.

    It Provides a Device selector and several dockWidgets for interacting and
    displaying information from the selected device.
    '''

    def __init__(self, parent=None, designMode=False):
        TaurusMainWindow.__init__(self, parent, designMode=designMode)
        self.loadUi()

        # setting up the device Tree.
        #@todo: This should be done in the ui file when the TaurusDatabaseTree Designer plugin is available
        import taurus.qt.qtgui.tree
        TaurusDbTreeWidget = taurus.qt.qtgui.tree.TaurusDbTreeWidget

        self.deviceTree = TaurusDbTreeWidget(
            perspective=TaurusElementType.Device)
        self.deviceTree.getQModel().setSelectables(
            [TaurusElementType.Member])  # TODO: Tango-centric
        # self.deviceTree.insertFilter(filterNonExported)
        self.setCentralWidget(self.deviceTree)

        # needed because of a limitation in when using the useParentModel
        # property from designer and taurus parents are not the same as Qt
        # Parents
        self._ui.taurusAttrForm.recheckTaurusParent()
        self._ui.taurusCommandsForm.recheckTaurusParent()

        # Add StateLed to statusBar
#        self.devStateLed = TaurusLed()
#        self.statusbar.addPermanentWidget(self.devStateLed)
#        self.devStateLed.setModel('/state')
#        self.devStateLed.setUseParentModel(True)

        # register subwidgets for configuration purposes
        # self.registerConfigDelegate(self.taurusAttrForm)
        # self.registerConfigDelegate(self.deviceTree)
        self.registerConfigDelegate(self._ui.taurusCommandsForm)

        self.loadSettings()
        self.createActions()

        # self.addToolBar(self.basicTaurusToolbar())

        self.deviceTree.currentItemChanged.connect(self.onItemSelectionChanged)

        self.updatePerspectivesMenu()
        if not designMode:
            self.splashScreen().finish(self)

    def createActions(self):
        '''create actions '''
        # View Menu
        self.showAttrAction = self.viewMenu.addAction(
            self._ui.attrDW.toggleViewAction())
        self.showCommandsAction = self.viewMenu.addAction(
            self._ui.commandsDW.toggleViewAction())
        self.showTrendAction = self.viewMenu.addAction(
            self._ui.trendDW.toggleViewAction())

    def setTangoHost(self, host):
        '''extended from :class:setTangoHost'''
        TaurusMainWindow.setTangoHost(self, host)
        self.deviceTree.setModel(host)
        # self.deviceTree.insertFilter(filterNonExported)

    def onItemSelectionChanged(self, current, previous):
        itemData = current.itemData()
        if isinstance(itemData, TangoDevInfo):  # TODO: Tango-centric
            self.onDeviceSelected(itemData)

    def onDeviceSelected(self, devinfo):
        devname = devinfo.name()
        msg = 'Connecting to "%s"...' % devname
        self.statusBar().showMessage(msg)
        # abort if the device is not exported
        if not devinfo.exported():  # TODO: Tango-centric
            msg = 'Connection to "%s" failed (not exported)' % devname
            self.statusBar().showMessage(msg)
            self.info(msg)
            Qt.QMessageBox.warning(self, "Device unreachable", msg)
            self.setModel('')
            return
        self.setDevice(devname)

    def setDevice(self, devname):
        # try to connect with the device
        self.setModel(devname)
        dev = self.getModelObj()
        state = dev.state
        # test the connection
        if state == TaurusDevState.Ready:
            msg = 'Connected to "%s"' % devname
            self.statusBar().showMessage(msg)
            self._ui.attrDW.setWindowTitle('Attributes - %s' % devname)
            self._ui.commandsDW.setWindowTitle('Commands - %s' % devname)
        else:
            # reset the model if the connection failed
            msg = 'Connection to "%s" failed (state = %s)' % (devname,
                                                              state.name)
            self.statusBar().showMessage(msg)
            self.info(msg)
            Qt.QMessageBox.warning(self, "Device unreachable", msg)
            self.setModel('')

    @classmethod
    def getQtDesignerPluginInfo(cls):
        ret = TaurusMainWindow.getQtDesignerPluginInfo()
        ret['module'] = 'taurus.qt.qtgui.panel'
        return ret

#=========================================================================
# Launchers
#=========================================================================


def TaurusDevicePanelMain():
    '''A launcher for TaurusDevicePanel.'''
    import sys
    from taurus.qt.qtgui.application import TaurusApplication
    from taurus.core.util import argparse

    parser = argparse.get_taurus_parser()
    parser.set_usage("%prog [options] [devname [attrs]]")
    parser.set_description("Taurus Application inspired in Jive and Atk Panel")
    parser.add_option("", "--config-file", dest="config_file", default=None,
                      help="load a config file (TODO: document this option)")

    app = TaurusApplication(cmd_line_parser=parser, app_name="TaurusDevicePanel",
                            app_version=taurus.Release.version)
    args = app.get_command_line_args()
    options = app.get_command_line_options()

    w = TaurusDevicePanel()
    w.show()

    if len(args) == 0:
        from taurus.qt.qtgui.panel import TaurusModelChooser
        models, ok = TaurusModelChooser.modelChooserDlg(w,
                                                        selectables=[
                                                            TaurusElementType.Member],
                                                        singleModel=True)
        model = models[0] if ok and models else None
        filters = ''
    else:
        model = args[0]
        filters = args[1:]

    if options.config_file is not None:
        w.loadConfigFile(options.config_file)
    elif model and filters:
        w.setAttributeFilters({model: filters})

    w.setModel(model)

    sys.exit(app.exec_())


def TaurusPanelMain():
    '''A launcher for TaurusPanel.'''
    # NOTE: DON'T PUT TEST CODE HERE.
    # THIS IS CALLED FROM THE LAUNCHER SCRIPT (<taurus>/scripts/tauruspanel)
    from taurus.qt.qtgui.application import TaurusApplication
    from taurus.core.util import argparse
    import sys

    parser = argparse.get_taurus_parser()
    parser.set_usage("%prog [options] [devname]")
    parser.set_description("Taurus Application inspired in Jive and Atk Panel")

    app = TaurusApplication(cmd_line_parser=parser, app_name="tauruspanel",
                            app_version=taurus.Release.version)
    args = app.get_command_line_args()
    options = app.get_command_line_options()

    w = TaurusDevPanel()

    if options.tango_host is None:
        options.tango_host = taurus.Authority().getNormalName()
    w.setTangoHost(options.tango_host)

    if len(args) == 1:
        w.setDevice(args[0])

    w.show()

    sys.exit(app.exec_())


###############################################################################

if __name__ == "__main__":
    TaurusDevicePanelMain()



#!/usr/bin/env python

#############################################################################
##
# This file is part of Taurus
##
# http://taurus-scada.org
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Taurus is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Taurus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
#############################################################################

"""Useful module for remote logging"""

from __future__ import print_function
from __future__ import with_statement

__all__ = ["LogRecordStreamHandler", "LogRecordSocketReceiver", "log"]

import time
import socket
import pickle
import logging
import logging.handlers
import struct
import weakref

try:
    import socketserver
except:
    import SocketServer as socketserver


class LogRecordStreamHandler(socketserver.StreamRequestHandler):

    def handle(self):
        try:
            return self._handle()
        except:
            pass

    def _handle(self):
        self._stop = stop = 0
        self.hostName = self.server.hostName
        self.server.registerHandler(self)
        while not stop:
            chunk = self.connection.recv(4)
            if len(chunk) < 4:
                break
            slen = struct.unpack('>L', chunk)[0]
            chunk = self.connection.recv(slen)
            while len(chunk) < slen:
                chunk = chunk + self.connection.recv(slen - len(chunk))
            obj = self.unPickle(chunk)
            record = self.makeLogRecord(obj)
            self.handleLogRecord(record)
            stop = self._stop

    def unPickle(self, data):
        return pickle.loads(data)

    def makeLogRecord(self, obj):
        record = logging.makeLogRecord(obj)
        if not hasattr(record, 'hostName'):
            record.hostName = self.hostName
        return record

    def handleLogRecord(self, record):
        logger = self.server.data.get("logger")
        if logger is None:
            logger = logging.getLogger(record.name)
        if not logger.isEnabledFor(record.levelno):
            return
        logger.handle(record)

    def stop(self):
        self._stop = 1


class LogRecordSocketReceiver(socketserver.ThreadingTCPServer):
    """
    Simple TCP socket-based logging receiver suitable for testing.
    """

    allow_reuse_address = 1
    daemon_threads = True

    def __init__(self, host='localhost',
                 port=logging.handlers.DEFAULT_TCP_LOGGING_PORT,
                 handler=LogRecordStreamHandler, **kwargs):
        socketserver.ThreadingTCPServer.__init__(self, (host, port), handler)
        self.hostName = socket.gethostbyaddr(host)[0]
        self.port = port
        self._stop = 0
        self._stopped = 0
        self.timeout = 1
        self.data = kwargs
        self.__handlers = []

    def registerHandler(self, handler):
        if handler is not None:
            self.__handlers.append(weakref.ref(handler))

    def unregisterHandler(self, handler):
        try:
            self.__handlers.remove(handler)
        except ValueError:
            pass

    def serve_until_stopped(self):
        import select
        stop = 0
        while not stop:
            rd, wr, ex = select.select([self.socket.fileno()],
                                       [], [],
                                       self.timeout)
            if rd:
                self.handle_request()
            stop = self._stop
        self._stopped = 1

    def stop(self):
        self._stop = True
        while not self._stopped:
            time.sleep(0.1)
        for handler in self.__handlers:
            h = handler()
            if h is not None:
                h.stop()
                h.finish()
                self.close_request(h.connection)
        self.socket.close()


class LogNameFilter(logging.Filter):

    def __init__(self, name=None):
        self.name = name

    def filter(self, record):
        name = self.name
        if name is None:
            return True
        return record.name == name


def log(host, port, name=None, level=None):
    local_logger_name = "RemoteLogger.%s.%d" % (host, port)
    local_logger = logging.getLogger(local_logger_name)

    if name is not None:
        local_logger.addFilter(LogNameFilter(name=name))

    if level is not None:
        local_logger.setLevel(level)

    tcpserver = LogRecordSocketReceiver(host=host, port=port,
                                        logger=local_logger)
    msg = "logging for '%s' on port %d" % (host, port)
    if name is not None:
        msg += " for " + name
    print("Start", msg)

    try:
        tcpserver.serve_until_stopped()
    except KeyboardInterrupt:
        print("\nCancelled", msg)


def main(argv=None):
    import optparse
    import socket

    import taurus.core.util.log

    taurus.setLogLevel(taurus.Trace)

    dft_port = logging.handlers.DEFAULT_TCP_LOGGING_PORT

    host = socket.gethostname()

    help_port = "port where log server is running [default: %d]" % dft_port
    help_name = "filter specific log object [default: None, meaning don't " \
                "filter]"
    help_level = "filter specific log level." \
                 "Allowed values are (case insensitive): critical, "\
                 "error, warning/warn, info, debug, trace [default: debug]."

    parser = optparse.OptionParser()
    parser.add_option("--log-port", dest="log_port", default=dft_port,
                      type="int", help=help_port)
    parser.add_option("--log-name", dest="log_name", default=None,
                      type="string", help=help_name)
    parser.add_option("--log-level", dest="log_level", default="debug",
                      type="string", help=help_level)

    if argv is None:
        import sys
        argv = sys.argv

    options, args = parser.parse_args(args=argv)

    port, name = options.log_port, options.log_name
    level_str = options.log_level.capitalize()

    level = None
    if hasattr(taurus, level_str):
        level = getattr(taurus, level_str)

    log(host, port, name=name, level=level)

if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-

##############################################################################
##
## This file is part of Taurus
##
## http://taurus-scada.org
##
## Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
## Taurus is free software: you can redistribute it and/or modify
## it under the terms of the GNU Lesser General Public License as published by
## the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.
##
## Taurus is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU Lesser General Public License for more details.
##
## You should have received a copy of the GNU Lesser General Public License
## along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
##############################################################################

"""This module exposes PyQt4/PyQt5/PySide module"""

__all__ = ["initialize", "getQtName", "getQt", "_updateQtSubModule", "requires"]


from taurus import tauruscustomsettings as __config
from taurus.core.util import log as __log

import os

__QT = None
__QT_NAME = None
__QT_KNOWN_APIS = "PyQt4", "PyQt5", "PySide"
__QT_PREFERED_APIS = None

__QT_INIT = False
__QT_LOG_INIT = False
__QT_RES_INIT = False


def __getPreferedAPIs():
    return [__config.QT_AUTO_API] + \
      [api for api in __QT_KNOWN_APIS if api != __config.QT_AUTO_API]


def __assertQt(name, qt=None, strict=True):
    qt = qt or __QT
    if name is None or qt is None:
        return
    qt_name = qt.__name__
    if qt_name != name:
        msg = "Cannot use %s because %s already in use" % (name, qt_name)
        if strict:
            raise Exception(msg)
        else:
            __log.error(msg)


def __hasBinding(qt_name):
    """Safely check for known qt apis without importing submodules"""
    import imp
    try:
        _, pth, _ = imp.find_module(qt_name)
        imp.find_module('QtCore', pth)
        imp.find_module('QtGui', pth)
    except ImportError:
        return False
    return True


def __import(name):
    import sys
    __import__(name)
    return sys.modules[name]


def __importQt(name):
    return __import(getQtName() + "." + name)


def __initializeQtLogging():
    QtCore = __importQt("QtCore")

    QT_LEVEL_MATCHER = {
        QtCore.QtDebugMsg:     __log.debug,
        QtCore.QtWarningMsg:   __log.warning,
        QtCore.QtCriticalMsg:  __log.critical,
        QtCore.QtFatalMsg:     __log.fatal,
        QtCore.QtSystemMsg:    __log.critical,
    }

    if hasattr(QtCore, "qInstallMessageHandler"):
        def taurusMessageHandler(msg_type, log_ctx, msg):
            f = QT_LEVEL_MATCHER.get(msg_type)
            return f("Qt%s %s.%s[%s]: %a", log_ctx.category, log_ctx.file,
                     log_ctx.function, log_ctx.line, msg)
        QtCore.qInstallMessageHandler(taurusMessageHandler)
    elif hasattr(QtCore, "qInstallMsgHandler"):
        def taurusMsgHandler(msg_type, msg):
            f = QT_LEVEL_MATCHER.get(msg_type)
            return f("Qt: " + msg)
        QtCore.qInstallMsgHandler(taurusMsgHandler)


def __getTheme():
    QtGui = __importQt("QtGui")
    return QtGui.QIcon.themeName()


def __hasTheme():
    return len(__getTheme()) > 0


def __get_taurus_resource_path():
    import os
    this_dir = os.path.abspath(os.path.dirname(__file__))
    taurus_dir = os.path.join(this_dir, os.path.pardir, os.path.pardir)
    base_resource_dir = os.path.join(taurus_dir, "qt", "qtgui", "resource")
    return os.path.realpath(base_resource_dir)


def __get_taurus_tango_theme_path():
    import os
    return os.path.join(__get_taurus_resource_path(), "tango-icons")


def __themeDirectories():
    """Returns valid theme directories for the current theme
    **Requires QApplication to be created**"""
    import os
    QtGui = __importQt("QtGui")
    theme = __getTheme()
    theme_paths = QtGui.QIcon.themeSearchPaths()
    result = []
    for theme_path in theme_paths:
        theme_path = os.path.join(theme_path, theme)
        if os.path.isdir(theme_path) and not theme_path in result:
            result.append(theme_path)
    return result


def __initializeTheme():
    """Currently not used"""
    import os
    QtGui = __importQt("QtGui")

    # Can only resources if QApplication already exists
    app = QtGui.QApplication.instance()
    if app is None:
        raise SystemError("QApplication object must exist before "
                          "initializing Qt resources")

    tango_theme_dir = __get_taurus_tango_theme_path()

    # initialize theme if necessary
    theme = __getTheme()
    has_theme = len(theme) > 0
    if not has_theme:
        __log.info("No native theme support. Using local tango-theme-icons")
        if os.path.isdir(tango_theme_dir):
            # If themes are not supported (windows, for example), taurus
            # initializes local Tango theme
            theme_search_path = QtGui.QIcon.themeSearchPaths()
            theme_search_path.append(tango_theme_dir)
            QtGui.QIcon.setThemeSearchPaths(theme_search_path)
            QtGui.QIcon.setThemeName("Tango")
        else:
            __log.warning("Local tango-theme-icons not found: Theme not initialized")


def __initializeQtResources():
    import os
    QtCore = __importQt("QtCore")
    base_resource_dir = __get_taurus_resource_path()

    # add taurus resources
    namespace = __config.NAMESPACE
    search_paths = QtCore.QDir.searchPaths(namespace) or []
    for elem in os.listdir(base_resource_dir):
        abs_elem = os.path.join(base_resource_dir, elem)
        if os.path.isdir(abs_elem) and not abs_elem in search_paths:
            search_paths.append(abs_elem)
    QtCore.QDir.setSearchPaths(namespace, search_paths)


def __removePyQtInputHook():
    try:
        __importQt("QtCore").pyqtRemoveInputHook()
    except AttributeError:
        pass


def _updateQtSubModule(glob_dict, qt_sub_module_name):
    glob_dict.update(__importQt(qt_sub_module_name).__dict__)


#------------------------------------------------------------------------------
# PyQt4
#------------------------------------------------------------------------------

def __get_sip():
    try:
        import sip
    except ImportError:
        sip = None
    return sip


def __setPyQt4API(element, api_version=2):
    sip = __get_sip()
    try:
        ver = sip.getapi(element)
    except ValueError:
        ver = -1

    if ver < 0:
        try:
            sip.setapi(element, api_version)
            __log.debug("%s API set to version %d",
                      element, sip.getapi(element))
        except ValueError:
            __log.warning("Error setting %s API to version %s", element,
                        api_version, exc_info=1)
            return False
    elif ver < api_version:
        __log.info("%s API set to version %s (advised: version >= %s)",
                 element, ver, api_version)
    return True


def __preparePyQt4():
    import sys

    # In python 3 APIs are set to level 2 so nothing to do
    if sys.version_info[0] > 2:
        return __import("PyQt4")

    sip = __get_sip()

    # For PySide compatibility, use the new-style string API that
    # automatically converts QStrings to Unicode Python strings. Also,
    # automatically unpack QVariants to their underlying objects.
    if sip is None:
        __log.warning("Could not find sip")
    elif sip.SIP_VERSION < 0x040900:
        sip_ver = sip.SIP_VERSION_STR
        __log.warning("Using old sip %s (advised >= 4.9)", sip_ver)
    else:
        for obj in ("QDate", "QDateTime", "QString", "QTextStream", "QTime",
          "QUrl", "QVariant"):
            __setPyQt4API(obj, 2)

    return __import("PyQt4")


#------------------------------------------------------------------------------
# PyQt5
#------------------------------------------------------------------------------

def __preparePyQt5():
    return __import("PyQt5")


#------------------------------------------------------------------------------
# PySide
#------------------------------------------------------------------------------

def __preparePySide():
    return __import("PySide")


#------------------------------------------------------------------------------
# Global
#------------------------------------------------------------------------------

def getQt(name=None, strict=True):
    global __QT, __QT_NAME
    if __QT:
        __assertQt(name, qt=__QT, strict=strict)
        return __QT

    import sys

    modules = sys.modules
    for api_name in __QT_PREFERED_APIS:
        api = modules.get(api_name)
        if api:
            __assertQt(name, qt=api, strict=strict)
            __QT = api
            __QT_NAME = name
            return __QT

    # no qt imported yet
    if strict and name:
        apis = [name]
    else:
        apis = list(__QT_PREFERED_APIS)
        if name:
            apis.remove(name)
            apis.insert(0, name)
    for api_name in apis:
        f = globals()["__prepare" + api_name]
        try:
            __QT = f()
            __QT_NAME = api_name
            return __QT
        except ImportError:
            continue
    raise ImportError("No suitable Qt found")


def getQtName(name=None, strict=True):
    # force initialization of Qt
    getQt(name=name, strict=strict)
    return __QT_NAME


def initialize(name=None, strict=True, logging=True,
               resources=True, remove_inputhook=True):
    global __QT_INIT

    if __QT_INIT:
        return getQt()

    if not __config.QT_AUTO_API in __QT_KNOWN_APIS:
        raise ImportError("Invalid QT_AUTO_API '%s'. Valid APIs are %s" % \
                          (__config.QT_AUTO_API,
                          ", ".join(__QT_KNOWN_APIS)))

    global __QT_PREFERED_APIS
    __QT_PREFERED_APIS = __getPreferedAPIs()

    qt = getQt(name=name, strict=strict)
    if logging:
        __initializeQtLogging()
    if resources:
        __initializeQtResources()
    if remove_inputhook:
        __removePyQtInputHook()

    __QT_INIT = True

    QT_API = os.environ.get('QT_API')
    if QT_API is None:
        global __QT_NAME
        if __QT_NAME == 'PySide':
            QT_API = 'pyside'
        else:
            QT_API = 'pyqt'

    os.environ['QT_API'] = QT_API

    return qt


def requires(origin=None, exc_class=ImportError, **kwargs):
    """
    Determines if the Qt component fulfills the minimum specified.
    Can take one of the following keyword arguments: Qt, PyQt, PySide.
    Any of these arguments maybe a string in the loose version format
    (see :class:`distutils.version.LooseVersion`)

    If present, *Qt* keyword stands for the minimum Qt C++ version.
    If present, *PyQt* keyword stands for the minimum PyQt version.
    If present, *PySide* keyword stands for the minimum PySide version.

    If a keyword is not present, it means it accepts any version. So, if, for
    example, you are running taurus with PySide and you call requires with
    `requires(PyQt='4.10')` it will **not** fail.

    :param Qt:
    """
    from distutils.version import LooseVersion as V
    QtName = getQtName()
    QtCore = __importQt("QtCore")

    if origin is None:
        msg_prefix = "Required"
    else:
        msg_prefix = origin + " requires"

    # check C++ Qt minimum version
    req_cpp_qt_v_str = kwargs.pop("Qt", None)
    if req_cpp_qt_v_str is not None:
        cpp_qt_v_str = QtCore.QT_VERSION_STR
        if V(req_cpp_qt_v_str) > V(cpp_qt_v_str):
            if exc_class:
                msg = "{0} C++ Qt >= {1}. Installed C++ Qt is {2}".format(
                    msg_prefix, req_cpp_qt_v_str, cpp_qt_v_str)
                raise exc_class(msg)
            else:
                return False

    if QtName.startswith('PyQt'):
        req_v_str = kwargs.get('PyQt', "0")
        qt_v_str = QtCore.PYQT_VERSION_STR
    elif QtName == 'PySide':
        req_v_str = kwargs.get('PySide', "0")
        qt_v_str = QtCore.PYSIDE_VERSION_STR

    if req_v_str:
        if V(req_v_str) > V(qt_v_str):
            if exc_class:
                msg = "{0} {1} >= {2}. Installed {1} is {3}",format(
                    msg_prefix, QtName, req_v_str, qt_v_str)
                raise exc_class(msg)
            else:
                return False
    return True


if __config.QT_AUTO_INIT:
    initialize(name=__config.QT_AUTO_API,
               strict=__config.QT_AUTO_STRICT,
               logging=__config.QT_AUTO_INIT_LOG,
               resources=__config.QT_AUTO_INIT_RES,
               remove_inputhook=__config.QT_AUTO_REMOVE_INPUTHOOK)

    __log.info('Using "%s" for Qt', __QT_NAME)
